import asyncio
from datetime import date, timedelta
from typing import Dict, Tuple

import pandas as pd
from pytz import timezone

from liualgotrader.common.database import fetch_as_dataframe
from liualgotrader.common.tlog import tlog

est = timezone("America/New_York")


def portfolio_return(
    env: str, start_date: date
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    df = load_trades_for_period(
        env, start_date, date.today() + timedelta(days=1)
    )

    # day X strategy = total return
    table: Dict = {}
    invested_table: Dict = {}
    percentages_table: Dict = {}
    for index, row in df.iterrows():
        d = pd.Timestamp(pd.to_datetime(row["client_time"]).date(), tz=est)
        if d not in table:
            table[d] = {}
            invested_table[d] = {}
            percentages_table[d] = {}
            table[d]["revenue"] = 0.0
            invested_table[d]["invested"] = 0.0

        strat = row["algo_name"]
        if not strat in table[d]:
            table[d][strat] = 0.0
            invested_table[d][strat] = 0.0
            percentages_table[d][strat] = 0.0
        delta = (
            (1.0 if row["operation"] == "sell" and row["qty"] > 0 else -1.0)
            * float(row["price"])
            * row["qty"]
        )
        table[d][strat] += delta
        table[d]["revenue"] += delta
        if row["operation"] == "buy":
            invested_table[d][strat] += row["qty"] * float(row["price"])
            invested_table[d]["invested"] += row["qty"] * float(row["price"])
        percentages_table[d][strat] = (
            (100.0 * table[d][strat] / invested_table[d][strat])
            if invested_table[d][strat] != 0
            else 0.0
        )
        percentages_table[d]["revenue"] = (
            (100.0 * table[d]["revenue"] / invested_table[d]["invested"])
            if invested_table[d][strat] != 0
            else 0.0
        )

    return (
        pd.DataFrame.from_dict(table, orient="index").sort_index(),
        pd.DataFrame.from_dict(invested_table, orient="index").sort_index(),
        pd.DataFrame.from_dict(percentages_table, orient="index").sort_index(),
    )


def load_trades_for_period(
    env: str, from_date: date, to_date: date
) -> pd.DataFrame:
    query = f"""
    SELECT client_time, symbol, operation, qty, price, algo_name
    FROM 
        new_trades as t, algo_run as a
    WHERE 
        t.algo_run_id = a.algo_run_id AND 
        t.tstamp >= '{from_date}' AND 
        t.tstamp < '{to_date}' AND
        t.expire_tstamp is null AND
        a.algo_env = '{env}'
    ORDER BY symbol, tstamp
    """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(fetch_as_dataframe(query))


def load_trades(day: date, env: str, end_date: date = None) -> pd.DataFrame:
    query = f"""
    SELECT t.*, a.batch_id, a.algo_name
    FROM 
    new_trades as t, algo_run as a
    WHERE 
        t.algo_run_id = a.algo_run_id AND 
        t.tstamp >= '{day}' AND 
        t.tstamp < '{day + timedelta(days=1) if not end_date else end_date}' AND
        t.expire_tstamp is null AND
        a.algo_env = '{env}'
    ORDER BY symbol, tstamp
    """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(fetch_as_dataframe(query))


async def aload_trades_by_batch_id(batch_id: str) -> pd.DataFrame:
    query = f"""
        SELECT 
            t.*, a.batch_id, a.start_time, a.algo_name
        FROM 
            new_trades as t, algo_run as a
        WHERE 
            t.algo_run_id = a.algo_run_id AND 
            a.batch_id = '{batch_id}' AND
            t.expire_tstamp is null 
        ORDER BY symbol, tstamp
    """
    df: pd.DataFrame = await fetch_as_dataframe(query)
    try:
        df["client_time"] = pd.to_datetime(df["client_time"])
    except Exception:
        tlog(
            f"[Error] aload_trades_by_batch_id({batch_id}) can't convert 'client_time' column to datetime"
        )
    return df


def load_trades_by_batch_id(batch_id: str) -> pd.DataFrame:
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(aload_trades_by_batch_id(batch_id))


def load_runs(day: date, env: str, end_date: date = None) -> pd.DataFrame:
    query = f"""
     SELECT * 
     FROM 
     algo_run as t
     WHERE 
         start_time >= '{day}' AND 
         start_time < '{day + timedelta(days=1) if not end_date else end_date}' AND
         algo_env = '{env}'
     ORDER BY start_time
     """
    loop = asyncio.get_event_loop()
    df = loop.run_until_complete(fetch_as_dataframe(query))
    df.set_index("algo_run_id", inplace=True)
    return df


def load_batch_list(day: date, env: str) -> pd.DataFrame:
    query = f"""
        SELECT DISTINCT a.batch_id, a.start_time
        FROM 
        new_trades as t, algo_run as a
        WHERE 
            t.algo_run_id = a.algo_run_id AND 
            t.tstamp >= '{day}' AND 
            t.tstamp < '{day + timedelta(days=1)}'AND
            t.expire_tstamp is null AND
            a.algo_env = '{env}'
        """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(fetch_as_dataframe(query))


def load_traded_symbols(batch_id: str) -> pd.DataFrame:
    query = f"""
        SELECT 
            DISTINCT t.symbol
        FROM 
            new_trades as t, algo_run as a
        WHERE 
            t.algo_run_id = a.algo_run_id AND 
            a.batch_id = '{batch_id}'
    """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(fetch_as_dataframe(query))


def load_batch_symbols(batch_id: str) -> pd.DataFrame:
    query = f"""
        SELECT 
            symbol
        FROM 
            trending_tickers
        WHERE 
            batch_id = '{batch_id}'
    """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(fetch_as_dataframe(query))


def calc_batch_revenue(
    symbol: str, trades: pd.DataFrame, batch_id: str = None
) -> float:
    symbol_df = trades[
        (trades["symbol"] == symbol)
        & (not batch_id or trades["batch_id"] == batch_id)
    ]
    rc = 0
    for index, row in symbol_df.iterrows():
        rc += (
            row["price"]
            * row["qty"]
            * (1 if row["operation"] == "sell" and row["qty"] > 0 else -1)
        )
    return round(rc, 2)


def calc_revenue(symbol: str, trades: pd.DataFrame, env) -> float:
    symbol_df = trades[
        (trades["symbol"] == symbol) & (trades["algo_env"] == env)
    ]
    rc = 0
    for index, row in symbol_df.iterrows():
        rc += (
            row["price"]
            * row["qty"]
            * (1 if row["operation"] == "sell" and row["qty"] > 0 else -1)
        )
    return round(rc, 2)


def count_trades(symbol, trades: pd.DataFrame, batch_id: str) -> int:
    symbol_df = trades[
        (trades["symbol"] == symbol) & (trades["batch_id"] == batch_id)
    ]
    return len(symbol_df.index)



import time
from concurrent.futures import ThreadPoolExecutor
from datetime import date, datetime, timedelta
from typing import Dict, List, Optional

import alpaca_trade_api as tradeapi
import pandas as pd
import requests
from alpaca_trade_api.common import get_polygon_credentials
from asyncpg.pool import Pool
from pandas import DataFrame as df
from pandas import Timestamp
from pytz import timezone

from liualgotrader.common import config, trading_data
from liualgotrader.common.decorators import timeit
from liualgotrader.common.tlog import tlog
from liualgotrader.fincalcs.vwap import add_daily_vwap
from liualgotrader.models.ticker_snapshot import TickerSnapshot

volume_today: Dict[str, int] = {}
minute_history: Dict[str, df] = {}
quotes: Dict[str, df] = {}


def get_historical_data_from_finnhub(symbols: List[str]) -> Dict[str, df]:

    tlog(f"Loading {len(symbols)} tickers historic data from Finnhub")
    nyc = timezone(NY := "America/New_York")
    _from = datetime.today().astimezone(nyc) - timedelta(days=30)
    _from = _from.replace(hour=9, minute=29, second=0, microsecond=0)
    _to = datetime.now(nyc)

    minute_history: Dict[str, df] = {}
    with requests.Session() as s:
        try:
            c = 0
            for symbol in symbols:
                retry = True
                while retry:
                    retry = False

                    url = f"{config.finnhub_base_url}/stock/candle?symbol={symbol}&resolution=1&from={_from.strftime('%s')}&to={_to.strftime('%s')}&token={config.finnhub_api_key}"
                    r = s.get(url)

                    if r.status_code == 200:
                        response = r.json()
                        if response["s"] != "no_data":
                            _data = {
                                "close": response["c"],
                                "open": response["o"],
                                "high": response["h"],
                                "low": response["l"],
                                "volume": response["v"],
                            }

                            _df = df(
                                _data,
                                index=[
                                    Timestamp(item, tz=NY, unit="s")
                                    for item in response["t"]
                                ],
                            )
                            c += 1
                            _df["vwap"] = 0.0
                            _df["average"] = 0.0
                            minute_history[symbol] = _df
                            tlog(
                                f"loaded {len(minute_history[symbol].index)} agg data points for {symbol} ({c}/{len(symbols)})"
                            )
                    elif r.status_code == 429:
                        tlog(
                            f"{symbols.index(symbol)}/{len(symbols)} API limit: ({r.text})"
                        )
                        time.sleep(30)
                        retry = True
                    else:
                        tlog(f"[ERROR] {r.status_code}, {r.text}")

        except KeyboardInterrupt:
            tlog("KeyboardInterrupt")
            pass

    return minute_history


def get_historical_data_from_poylgon_for_symbols(
    api: tradeapi, symbols: List[str], start_date: date, end_date: date
) -> Dict[str, df]:
    minute_history = {}
    for symbol in symbols:
        if symbol not in minute_history:
            minute_history[symbol] = api.polygon.historic_agg_v2(
                symbol,
                1,
                "minute",
                _from=start_date,
                to=end_date,
            ).df.tz_convert("US/Eastern")
            add_daily_vwap(minute_history[symbol])

    return minute_history


def get_historical_data_from_polygon_by_range(
    api: tradeapi, symbols: List[str], start_date: date, timespan: str
) -> Dict[str, df]:
    """get ticker history"""

    _minute_history: Dict[str, df] = {}
    try:
        for symbol in symbols:
            from_date = start_date
            while from_date < date.today():
                retry = 5
                _df = None
                while retry > 0:
                    try:
                        _df = api.polygon.historic_agg_v2(
                            symbol,
                            1,
                            timespan,
                            _from=str(from_date),
                            to=str(
                                from_date
                                + timedelta(
                                    days=1 + config.polygon.MAX_DAYS_TO_LOAD
                                )
                            ),
                        ).df
                        break

                    except Exception as e:

                        retry -= 1

                if _df is None or not len(_df):
                    break

                _df["vwap"] = 0.0
                _df["average"] = 0.0

                _minute_history[symbol] = (
                    pd.concat([_minute_history[symbol], _df])
                    if symbol in _minute_history
                    else _df
                )

                from_date = _df.index[-1] + timedelta(days=1)

                tlog(
                    f"get_historical_data_from_polygon_by_range() - total loaded {len(_minute_history[symbol].index)} agg data points for {symbol}"
                )
    except KeyboardInterrupt:
        tlog("KeyboardInterrupt")

    return _minute_history


def get_symbol_data(
    api: tradeapi, symbol: str, start_date: date, end_date: date
) -> Optional[df]:

    retry = 5
    df = None
    while retry > 0:
        try:
            df = api.polygon.historic_agg_v2(
                symbol,
                1,
                "minute",
                _from=str(start_date),
                to=str(end_date),
            ).df

            df["vwap"] = 0.0
            df["average"] = 0.0

            break
        except Exception as e:
            retry -= 1
            continue

    return df


def get_historical_daily_from_polygon_by_range(
    api: tradeapi, symbols: List[str], start_date: date, end_date: date
) -> Dict[str, df]:
    """get ticker history"""

    _minute_history: Dict[str, df] = {}
    try:
        for symbol in symbols:
            retry = 5
            _df = None
            while retry > 0:
                try:
                    _df = api.polygon.historic_agg_v2(
                        symbol,
                        1,
                        "day",
                        _from=str(start_date),
                        to=str(end_date),
                    ).df

                    _df["vwap"] = 0.0
                    _df["average"] = 0.0

                    _minute_history[symbol] = (
                        pd.concat([_minute_history[symbol], _df])
                        if symbol in _minute_history
                        else _df
                    )
                    break
                except Exception as e:
                    retry -= 1
                    continue

    except KeyboardInterrupt:
        tlog("KeyboardInterrupt")

    return _minute_history


def get_historical_data_from_polygon(
    api: tradeapi, symbols: List[str], max_tickers: int
) -> Dict[str, df]:
    """get ticker history"""

    tlog(f"Loading max {max_tickers} tickers w/ highest volume from Polygon")
    minute_history: Dict[str, df] = {}
    c = 0
    exclude_symbols = []
    try:
        for symbol in symbols:
            if symbol not in minute_history:
                retry_counter = 5
                while retry_counter > 0:
                    try:
                        if c < max_tickers:
                            _df = api.polygon.historic_agg_v2(
                                symbol,
                                1,
                                "minute",
                                _from=str(date.today() - timedelta(days=10)),
                                to=str(date.today() + timedelta(days=1)),
                            ).df
                            _df["vwap"] = 0.0
                            _df["average"] = 0.0

                            minute_history[symbol] = _df
                            tlog(
                                f"loaded {len(minute_history[symbol].index)} agg data points for {symbol} {c+1}/{max_tickers}"
                            )
                            c += 1
                            break

                        exclude_symbols.append(symbol)
                        break
                    except (
                        requests.exceptions.HTTPError,
                        requests.exceptions.ConnectionError,
                    ):
                        retry_counter -= 1
                        if retry_counter == 0:
                            exclude_symbols.append(symbol)
    except KeyboardInterrupt:
        tlog("KeyboardInterrupt")

    for x in exclude_symbols:
        symbols.remove(x)

    tlog(f"Total number of symbols for trading {len(symbols)}")
    return minute_history


@timeit
async def calculate_trends(pool: Pool) -> bool:
    # load snapshot
    with requests.Session() as session:
        url = (
            "https://api.polygon.io/"
            + "v2/snapshot/locale/us/markets/stocks/tickers"
        )
        with session.get(
            url,
            params={"apiKey": get_polygon_credentials(config.prod_api_key_id)},
        ) as response:
            if (
                response.status_code == 200
                and (r := response.json())["status"] == "OK"
            ):
                for ticker in r["tickers"]:
                    trading_data.snapshot[ticker.ticker] = TickerSnapshot(
                        symbol=ticker["ticker"],
                        volume=ticker["day"]["volume"],
                        today_change=ticker["todaysChangePerc"],
                    )

                if not trading_data.snapshot:
                    tlog("calculate_trends(): market snapshot not available")
                    return False

                # calculate sector trends
                sectors = await get_market_industries(pool)

                sector_tickers = {}
                for sector in sectors:
                    sector_tickers[sector] = await get_sector_tickers(
                        pool, sector
                    )

                    sector_volume = 0
                    adjusted_sum = 0.0
                    for symbol in sector_tickers[sector]:
                        sector_volume += trading_data.snapshot[symbol].volume
                        adjusted_sum += (
                            trading_data.snapshot[symbol].volume
                            * trading_data.snapshot[symbol].today_change
                        )

                    trading_data.sector_trend[sector] = round(
                        adjusted_sum / sector_volume, 2
                    )

                # calculate industry
                industries = await get_market_industries(pool)

                industry_tickers = {}
                for industry in industries:
                    industry_tickers[industry] = await get_industry_tickers(
                        pool, industry
                    )

                    industry_volume = 0
                    adjusted_sum = 0.0
                    for symbol in industry_tickers[sector]:
                        industry_volume += trading_data.snapshot[symbol].volume
                        adjusted_sum += (
                            trading_data.snapshot[symbol].volume
                            * trading_data.snapshot[symbol].today_change
                        )

                    trading_data.industry_trend[industry] = round(
                        adjusted_sum / industry_volume, 2
                    )

                return True

    return False


async def get_sector_tickers(pool: Pool, sector: str) -> List[str]:
    async with pool.acquire() as conn:
        async with conn.transaction():
            records = await conn.fetch(
                """
                    SELECT symbol
                    FROM ticker_data
                    WHERE sector = $1
                """,
                sector,
            )

            return [record[0] for record in records if record[0]]


async def get_industry_tickers(pool: Pool, industry: str) -> List[str]:
    async with pool.acquire() as conn:
        async with conn.transaction():
            records = await conn.fetch(
                """
                    SELECT symbol
                    FROM ticker_data
                    WHERE industry = $1
                """,
                industry,
            )

            return [record[0] for record in records if record[0]]


async def get_market_sectors(pool: Pool) -> List[str]:
    async with pool.acquire() as conn:
        async with conn.transaction():
            records = await conn.fetch(
                """
                    SELECT DISTINCT sector
                    FROM ticker_data
                """
            )
            return [record[0] for record in records if record[0]]


async def get_market_industries(pool: Pool) -> List[str]:
    async with pool.acquire() as conn:
        async with conn.transaction():
            records = await conn.fetch(
                """
                    SELECT DISTINCT industry
                    FROM ticker_data
                """
            )

            return [record[0] for record in records if record[0]]



import asyncio
import json
from multiprocessing import Queue
from typing import Awaitable, Dict, List

import websockets

from liualgotrader.common import market_data
from liualgotrader.common.tlog import tlog

from ..common import config
from .streaming_base import StreamingBase, WSConnectState

NY = "America/New_York"


class AlpacaStreaming(StreamingBase):
    END_POINT = "wss://data.alpaca.markets/stream"

    def __init__(self, key: str, secret: str, queues: List[Queue]):
        self.key = key
        self.secret = secret
        self.state: WSConnectState = WSConnectState.NOT_CONNECTED
        self.websocket: websockets.client.WebSocketClientProtocol
        self.consumer_task: asyncio.Task
        self.stream_map: Dict = {}
        super().__init__(queues)

    async def connect(self) -> bool:
        """Connect web-socket and authenticate, update internal state"""
        try:
            self.websocket = await websockets.client.connect(self.END_POINT)
            self.state = WSConnectState.CONNECTED
        except websockets.WebSocketException as wse:
            tlog(f"Exception when connecting to Alpaca WS {wse}")
            self.state = WSConnectState.NOT_CONNECTED
            return False

        auth_payload = {
            "action": "authenticate",
            "data": {"key_id": self.key, "secret_key": self.secret},
        }
        await self.websocket.send(json.dumps(auth_payload))
        _greeting = await self.websocket.recv()

        if isinstance(_greeting, bytes):
            _greeting = _greeting.decode("utf-8")
        msg = json.loads(_greeting)
        if msg.get("data", {}).get("status") != "authorized":
            tlog(f"Invalid Alpaca API credentials, Failed to authenticate: {msg}")
            raise ValueError(
                f"Invalid Alpaca API credentials, Failed to authenticate: {msg}"
            )

        self.state = WSConnectState.AUTHENTICATED

        self.consumer_task = asyncio.create_task(
            self._consumer(), name="alpaca-streaming-consumer-task"
        )

        tlog("Successfully connected to Alpaca web-socket")
        return True

    async def close(self) -> None:
        """Close open websocket, if open"""
        if self.state not in (
            WSConnectState.AUTHENTICATED,
            WSConnectState.CONNECTED,
        ):
            raise ValueError("can't close a non-open connection")
        try:
            await self.websocket.close()
        except websockets.WebSocketException as wse:
            tlog(f"failed to close web-socket w exception {wse}")

        self.state = WSConnectState.NOT_CONNECTED

    async def subscribe(self, symbol: str, handler: Awaitable) -> bool:
        if self.state != WSConnectState.AUTHENTICATED:
            raise ValueError(
                f"{symbol} web-socket not ready for listening, make sure connect passed successfully"
            )

        _subscribe_payload = {
            "action": "listen",
            "data": {"streams": [f"alpacadatav1/AM.{symbol}"]},
        }
        await self.websocket.send(json.dumps(_subscribe_payload))

        q_id = int(
            list(market_data.minute_history.keys()).index(symbol)
            / config.num_consumer_processes_ratio
        )
        self.stream_map[symbol] = (handler, q_id)
        return True

    async def unsubscribe(self, symbol: str) -> bool:
        if self.state != WSConnectState.AUTHENTICATED:
            raise ValueError(
                f"{symbol} web-socket not ready for listening, make sure connect passed successfully"
            )

        _subscribe_payload = {
            "action": "unlisten",
            "data": {"streams": [f"alpacadatav1/AM.{symbol}"]},
        }
        await self.websocket.send(json.dumps(_subscribe_payload))
        self.stream_map.pop(symbol, None)
        return False

    async def _reconnect(self) -> None:
        """automatically reconnect socket, and re-subscribe, internal"""
        tlog(f"{self.consumer_task.get_name()} reconnecting")
        await self.close()
        if await self.connect():
            _dict = self.stream_map.copy()
            self.stream_map.clear()

            for symbol in _dict:
                await self.subscribe(symbol, _dict[symbol])
        else:
            tlog(
                f"{self.consumer_task.get_name()} failed reconnect check log for reason"
            )

    async def _consumer(self) -> None:
        """Main tread loop for consuming incoming messages, internal only """

        tlog(f"{self.consumer_task.get_name()} starting")
        try:
            while True:
                _msg = await self.websocket.recv()
                if isinstance(_msg, bytes):
                    _msg = _msg.decode("utf-8")
                msg = json.loads(_msg)
                stream = msg.get("stream")
                if stream != "listening":
                    try:
                        _func, _q_id = self.stream_map.get(stream[3:], None)
                        if _func:
                            await _func(stream, msg["data"], self.queues[_q_id])
                        else:
                            tlog(
                                f"{self.consumer_task.get_name()} received {_msg} to an unknown stream {stream}"
                            )
                    except Exception as e:
                        tlog(
                            f"{self.consumer_task.get_name()}  exception {e.__class__.__qualname__}:{e}"
                        )

        except websockets.WebSocketException as wse:
            tlog(f"{self.consumer_task.get_name()} received WebSocketException {wse}")
            await self._reconnect()
        except asyncio.CancelledError:
            tlog(f"{self.consumer_task.get_name()} cancelled")

        tlog(f"{self.consumer_task.get_name()} completed")

    @classmethod
    async def minutes_handler(cls, symbol: str, data: Dict, queue: Queue) -> None:
        if data["ev"] != "AM":
            tlog(
                f"AlpacaStreaming.minutes_handler() got invalid event data: {symbol}:{data}"
            )
            return

        if symbol[3:] != data["T"]:
            tlog(
                f"AlpacaStreaming.minutes_handler() symbol does not match data payload {symbol}:{data}"
            )
            return

        try:
            data["EV"] = "AM"
            data["open"] = data["o"]
            data["high"] = data["h"]
            data["low"] = data["l"]
            data["close"] = data["c"]
            data["volume"] = data["v"]
            data["vwap"] = data["vw"]
            data["average"] = data["a"]
            queue.put(json.dumps(data))
        except Exception as e:
            tlog(
                f"Exception in handle_minute_bar(): exception of type {type(e).__name__} with args {e.args}"
            )



#!/usr/bin/env python

import asyncio
import importlib.util
import sys
import traceback
import uuid
from datetime import date, datetime, timedelta
from typing import Dict, List, Optional, Tuple

import alpaca_trade_api as tradeapi
import nest_asyncio
import pandas as pd
import pytz
from requests.exceptions import HTTPError
from tabulate import tabulate

from liualgotrader.analytics.analysis import load_trades_by_batch_id
from liualgotrader.common import config, market_data, trading_data
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.decorators import timeit
from liualgotrader.common.tlog import tlog
from liualgotrader.fincalcs.vwap import add_daily_vwap
from liualgotrader.models.algo_run import AlgoRun
from liualgotrader.models.new_trades import NewTrade
from liualgotrader.models.trending_tickers import TrendingTickers
from liualgotrader.scanners.base import Scanner
from liualgotrader.scanners.momentum import Momentum
from liualgotrader.strategies.base import Strategy, StrategyType


def get_batch_list():
    @timeit
    async def get_batch_list_worker():
        await create_db_connection()
        data = await AlgoRun.get_batches()
        print(
            tabulate(
                data,
                headers=["build", "batch_id", "strategy", "env", "start time"],
            )
        )

    try:
        if not asyncio.get_event_loop().is_closed():
            asyncio.get_event_loop().close()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop.run_until_complete(get_batch_list_worker())
    except KeyboardInterrupt:
        tlog("get_batch_list() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"get_batch_list() - exception of type {type(e).__name__} with args {e.args}"
        )
        traceback.print_exc()


"""
starting
"""


def show_usage():
    print(
        f"usage: {sys.argv[0]} --batch-list OR [--strict] [--symbol=SYMBOL] [--debug=SYMBOL] [--duration=<minutes>] <batch-id>\n"
    )
    msg = """
    'backter' application re-runs a past trading session, with new or modified
     strategies specified in tradeplan.toml. 'backter' application looks for 
     tradeplan.toml in current directory. The 'backter' application expects a 
     batch-id (UDID) as input. Using the --batch-list option you 
     can see a list of recent sessions to choose from. 
    """
    print(msg)
    print("options:")
    print(
        "--batch-list\tDisplay list of trading sessions, list limited to last 30 days"
    )
    print("--symbol\tRun on specific SYMBOL, bypass batch-id scanners")
    print(
        "--duration\tRun back-test for number of <minutes>, bypass batch-id run duration"
    )
    print(
        "--debug\tWrite verbose debug information for symbol SYMBOL during back-testing"
    )
    print(
        "--strict\tRun back-test session only on same symbols traded in the original batch"
    )


def show_version(filename: str, version: str) -> None:
    """Display welcome message"""
    print(f"filename:{filename}\ngit version:{version}\n")


async def create_strategies(
    conf_dict: Dict,
    duration: timedelta,
    ref_run_id: Optional[int],
    uid: str,
    start: datetime,
    bypass_strategy_duration: bool = False,
) -> None:
    strategy_types = []
    for strategy in conf_dict["strategies"]:
        print(strategy)
        strategy_name = strategy
        strategy_details = conf_dict["strategies"][strategy_name]
        if strategy_name == "MomentumLong":
            tlog(f"strategy {strategy_name} selected")
            strategy_types += [(strategy_details)]
        else:
            tlog(f"custom strategy {strategy_name} selected")

            try:
                spec = importlib.util.spec_from_file_location(
                    "module.name", strategy_details["filename"]
                )
                custom_strategy_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(custom_strategy_module)  # type: ignore
                class_name = strategy_name
                custom_strategy = getattr(custom_strategy_module, class_name)

                if not issubclass(custom_strategy, Strategy):
                    tlog(
                        f"custom strartegy must inherit from class {Strategy.__name__}"
                    )
                    exit(0)
                strategy_details.pop("filename", None)
                strategy_types += [(custom_strategy, strategy_details)]

            except Exception as e:
                tlog(
                    f"[Error]exception of type {type(e).__name__} with args {e.args}"
                )
                traceback.print_exc()
                exit(0)

    for strategy_tuple in strategy_types:
        strategy_type = strategy_tuple[0]
        strategy_details = strategy_tuple[1]
        config.env = "BACKTEST"
        tlog(f"initializing {strategy_type.name}")

        if "schedule" not in strategy_details:
            print("duration:", duration)
            strategy_details["schedule"] = [
                {
                    "start": int(
                        (
                            start - start.replace(hour=13, minute=30)
                        ).total_seconds()
                        // 60
                    ),
                    "duration": int(duration.total_seconds() // 60),
                }
            ]
        if bypass_strategy_duration:
            for schedule in strategy_details["schedule"]:
                schedule["duration"] = duration.total_seconds() // 60

        s = strategy_type(
            batch_id=uid, ref_run_id=ref_run_id, **strategy_details
        )
        await s.create()
        trading_data.strategies.append(s)


@timeit
async def backtest_symbol(
    data_api: tradeapi,
    portfolio_value: float,
    symbol: str,
    start: datetime,
    duration: timedelta,
    scanner_start_time: datetime,
    debug_symbol: bool = False,
) -> None:
    est = pytz.timezone("America/New_York")
    scanner_start_time = (
        pytz.utc.localize(scanner_start_time).astimezone(est)
        if scanner_start_time.tzinfo is None
        else scanner_start_time
    )
    start_time = pytz.utc.localize(start).astimezone(est)

    if scanner_start_time > start_time + duration:
        print(
            f"{symbol} picked too late at {scanner_start_time} ({start_time}, {duration})"
        )
        return

    start_time = scanner_start_time
    if start_time.second > 0:
        start_time = start_time.replace(second=0, microsecond=0)
    print(
        f"--> back-testing {symbol} from {str(start_time)} duration {duration}"
    )
    if debug_symbol:
        print("--> using DEBUG mode")

    re_try = 3

    while re_try > 0:
        # load historical data
        try:
            symbol_data = data_api.polygon.historic_agg_v2(
                symbol,
                1,
                "minute",
                _from=str(start_time - timedelta(days=8)),
                to=str(start_time + timedelta(days=1)),
                limit=10000,
            ).df
        except HTTPError as e:
            tlog(f"Received HTTP error {e} for {symbol}")
            return

        if len(symbol_data) < 100:
            tlog(f"not enough data-points  for {symbol}")
            return

        add_daily_vwap(
            symbol_data,
            debug=debug_symbol,
        )
        market_data.minute_history[symbol] = symbol_data
        print(
            f"loaded {len(market_data.minute_history[symbol].index)} agg data points"
        )
        try:
            minute_index = symbol_data["close"].index.get_loc(
                start_time, method="nearest"
            )
            break
        except (Exception, ValueError) as e:
            print(f"[EXCEPTION] {e} - trying to reload-data. ")
            re_try -= 1

    if re_try <= 0:
        return

    position: int = 0
    new_now = symbol_data.index[minute_index]
    print(f"start time with data {new_now}")
    price = 0.0
    last_run_id = None
    # start_time + duration
    while (
        new_now < config.market_close
        and minute_index < symbol_data.index.size - 1
    ):
        if symbol_data.index[minute_index] != new_now:
            print("mismatch!", symbol_data.index[minute_index], new_now)
            print(symbol_data["close"][minute_index - 10 : minute_index + 1])
            raise Exception()

        price = symbol_data["close"][minute_index]
        for strategy in trading_data.strategies:
            if debug_symbol:
                print(
                    f"Execute strategy {strategy.name} on {symbol} at {new_now}"
                )

            try:
                do, what = await strategy.run(
                    symbol,
                    True,
                    position,
                    symbol_data[: minute_index + 1],
                    new_now,
                    portfolio_value,
                    debug=debug_symbol,  # type: ignore
                    backtesting=True,
                )
            except Exception as e:
                tlog(
                    f"[ERROR] exception {e} on symbol {symbol} @ {strategy.name}"
                )
                continue

            if do:
                if (
                    what["side"] == "buy"
                    and float(what["qty"]) > 0
                    or what["side"] == "sell"
                    and float(what["qty"]) < 0
                ):
                    position += int(float(what["qty"]))
                    trading_data.buy_time[symbol] = new_now.replace(
                        second=0, microsecond=0
                    )
                else:
                    position -= int(float(what["qty"]))

                trading_data.last_used_strategy[symbol] = strategy

                db_trade = NewTrade(
                    algo_run_id=strategy.algo_run.run_id,
                    symbol=symbol,
                    qty=int(float(what["qty"])),
                    operation=what["side"],
                    price=price,
                    indicators=trading_data.buy_indicators[symbol]
                    if what["side"] == "buy"
                    else trading_data.sell_indicators[symbol],
                )

                await db_trade.save(
                    config.db_conn_pool,
                    str(new_now),
                    trading_data.stop_prices[symbol],
                    trading_data.target_prices[symbol],
                )

                if what["side"] == "buy":
                    await strategy.buy_callback(
                        symbol, price, int(float(what["qty"]))
                    )
                    break
                elif what["side"] == "sell":
                    await strategy.sell_callback(
                        symbol, price, int(float(what["qty"]))
                    )
                    break
            last_run_id = strategy.algo_run.run_id

        minute_index += 1
        new_now = symbol_data.index[minute_index]

    if position:
        if (
            trading_data.last_used_strategy[symbol].type
            == StrategyType.DAY_TRADE
        ):
            tlog(f"[{new_now}]{symbol} liquidate {position} at {price}")
            db_trade = NewTrade(
                algo_run_id=last_run_id,  # type: ignore
                symbol=symbol,
                qty=int(position) if int(position) > 0 else -int(position),
                operation="sell" if position > 0 else "buy",
                price=price,
                indicators={"liquidate": 1},
            )
            await db_trade.save(
                config.db_conn_pool,
                str(symbol_data.index[minute_index - 1]),
            )


def backtest(
    batch_id: str,
    conf_dict: Dict,
    debug_symbols: List[str],
    strict: bool = False,
    specific_symbols: List[str] = None,
    bypass_duration: int = None,
) -> str:
    data_api: tradeapi = tradeapi.REST(
        base_url=config.prod_base_url,
        key_id=config.prod_api_key_id,
        secret_key=config.prod_api_secret,
    )
    portfolio_value: float = (
        100000.0 if not config.portfolio_value else config.portfolio_value
    )
    uid = str(uuid.uuid4())

    async def backtest_run(
        start: datetime,
        duration: timedelta,
        ref_run_id: int,
        specific_symbols: List[str] = None,
    ) -> None:
        if specific_symbols:
            symbols_and_start_time: List = []
            for symbol in specific_symbols:
                symbols_and_start_time.append((symbol, start))
            num_symbols = len(specific_symbols)
        elif not strict:
            symbols_and_start_time = await TrendingTickers.load(batch_id)

            num_symbols = len(symbols_and_start_time)
        else:
            print("strict mode selected, loading symbols from trades")
            nest_asyncio.apply()
            _df = load_trades_by_batch_id(batch_id)
            symbols = _df.symbol.unique().tolist()
            num_symbols = len(symbols)
            est = pytz.timezone("America/New_York")
            start_time = pytz.utc.localize(_df.start_time.min()).astimezone(
                est
            )
            symbols_and_start_time = list(
                zip(symbols, [start_time for x in range(num_symbols)])
            )

        print(f"loaded {len(symbols_and_start_time)} symbols")

        if num_symbols > 0:
            est = pytz.timezone("America/New_York")
            start_time = pytz.utc.localize(start).astimezone(est)
            config.market_open = start_time.replace(
                hour=9, minute=30, second=0, microsecond=0
            )
            config.market_close = start_time.replace(
                hour=16, minute=0, second=0, microsecond=0
            )
            print(f"market_open {config.market_open}")
            await create_strategies(
                conf_dict,
                duration,
                ref_run_id,
                uid,
                start,
                bypass_duration is not None,
            )

            for symbol_and_start_time in symbols_and_start_time:
                symbol = symbol_and_start_time[0]
                await backtest_symbol(
                    data_api=data_api,
                    portfolio_value=portfolio_value,
                    symbol=symbol,
                    start=start,
                    duration=duration,
                    scanner_start_time=symbol_and_start_time[1],
                    debug_symbol=True if symbol in debug_symbols else False,
                )

    @timeit
    async def backtest_worker() -> None:
        await create_db_connection()
        run_details = await AlgoRun.get_batch_details(batch_id)
        run_ids, starts, ends, _ = zip(*run_details)

        if not len(run_details):
            print(f"can't load data for batch id {batch_id}")
        else:
            await backtest_run(
                start=min(starts),
                duration=timedelta(
                    minutes=max(
                        [
                            w["duration"]
                            for w in [
                                item
                                for sublist in [
                                    conf_dict["strategies"][s]["schedule"]  # type: ignore
                                    for s in conf_dict["strategies"]  # type: ignore
                                ]
                                for item in sublist
                            ]
                        ]
                    )
                    if not bypass_duration
                    else bypass_duration
                ),
                ref_run_id=run_ids[0],
                specific_symbols=specific_symbols,
            )

    try:
        if not asyncio.get_event_loop().is_closed():
            asyncio.get_event_loop().close()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop.run_until_complete(backtest_worker())
    except KeyboardInterrupt:
        tlog("backtest() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"backtest() - exception of type {type(e).__name__} with args {e.args}"
        )
        traceback.print_exc()
    finally:
        print("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=")
        print(f"new batch-id: {uid}")
        return uid


class BackTestDay:
    def __init__(self, conf_dict: Dict):
        self.uid = str(uuid.uuid4())

        self.data_api: tradeapi = tradeapi.REST(
            base_url=config.prod_base_url,
            key_id=config.prod_api_key_id,
            secret_key=config.prod_api_secret,
        )

        self.conf_dict = conf_dict
        config.portfolio_value = self.conf_dict.get("portfolio_value", None)
        self.minute_history: Dict[str, pd.DataFrame] = {}
        self.scanners: List[Scanner] = []

    async def create(self, day: date) -> str:
        await create_db_connection()
        scanners_conf = self.conf_dict["scanners"]

        est = pytz.timezone("America/New_York")
        start_time = datetime.combine(day, datetime.min.time()).astimezone(est)
        day = datetime.combine(day, datetime.min.time()).astimezone(est)
        self.start = day.replace(hour=9, minute=30)
        self.end = day.replace(hour=16, minute=0)

        config.market_open = start_time.replace(
            hour=9, minute=30, second=0, microsecond=0
        )
        config.market_close = start_time.replace(
            hour=16, minute=0, second=0, microsecond=0
        )
        for scanner_name in scanners_conf:
            scanner_object: Optional[Scanner] = None
            if scanner_name == "momentum":
                scanner_details = scanners_conf[scanner_name]
                try:
                    recurrence = scanner_details.get("recurrence", None)
                    target_strategy_name = scanner_details.get(
                        "target_strategy_name", None
                    )
                    scanner_object = Momentum(
                        provider=scanner_details["provider"],
                        data_api=self.data_api,
                        min_last_dv=scanner_details["min_last_dv"],
                        min_share_price=scanner_details["min_share_price"],
                        max_share_price=scanner_details["max_share_price"],
                        min_volume=scanner_details["min_volume"],
                        from_market_open=scanner_details["from_market_open"],
                        today_change_percent=scanner_details["min_gap"],
                        recurrence=timedelta(minutes=recurrence)
                        if recurrence
                        else None,
                        target_strategy_name=target_strategy_name,
                        max_symbols=scanner_details.get(
                            "max_symbols", config.total_tickers
                        ),
                    )
                    tlog(f"instantiated momentum scanner")
                except KeyError as e:
                    tlog(
                        f"Error {e} in processing of scanner configuration {scanner_details}"
                    )
                    exit(0)
            else:
                tlog(f"custom scanner {scanner_name} selected")
                scanner_details = scanners_conf[scanner_name]
                try:
                    spec = importlib.util.spec_from_file_location(
                        "module.name", scanner_details["filename"]
                    )
                    custom_scanner_module = importlib.util.module_from_spec(
                        spec
                    )
                    spec.loader.exec_module(custom_scanner_module)  # type: ignore
                    class_name = scanner_name
                    custom_scanner = getattr(custom_scanner_module, class_name)

                    if not issubclass(custom_scanner, Scanner):
                        tlog(
                            f"custom scanner must inherit from class {Scanner.__name__}"
                        )
                        exit(0)

                    scanner_details.pop("filename")
                    if "recurrence" not in scanner_details:
                        scanner_object = custom_scanner(
                            data_api=self.data_api,
                            **scanner_details,
                        )
                    else:
                        recurrence = scanner_details.pop("recurrence")
                        scanner_object = custom_scanner(
                            data_api=self.data_api,
                            recurrence=timedelta(minutes=recurrence),
                            **scanner_details,
                        )

                except Exception as e:
                    tlog(
                        f"[Error] scanners_runner.scanners_runner() for {scanner_name}:{e} "
                    )
            if scanner_object:
                self.scanners.append(scanner_object)

        await create_strategies(
            self.conf_dict,
            self.end - self.start,
            None,
            self.uid,
            day.replace(hour=9, minute=30, second=0, microsecond=0),
        )

        self.now = pd.Timestamp(self.start)
        self.symbols: List = []
        self.portfolio_value: float = (
            100000.0 if not config.portfolio_value else config.portfolio_value
        )
        if "risk" in self.conf_dict:
            config.risk = self.conf_dict["risk"]
        return self.uid

    async def next_minute(self) -> Tuple[bool, List[Optional[str]]]:
        rc_msg: List[Optional[str]] = []
        if self.now < self.end:
            for i in range(0, len(self.scanners)):
                if self.now == self.start or (
                    self.scanners[i].recurrence is not None
                    and self.scanners[i].recurrence.total_seconds() > 0  # type: ignore
                    and int((self.now - self.start).total_seconds() // 60)  # type: ignore
                    % int(self.scanners[i].recurrence.total_seconds() // 60)  # type: ignore
                    == 0
                ):
                    new_symbols = await self.scanners[i].run(self.now)
                    if new_symbols:
                        really_new = [
                            x for x in new_symbols if x not in self.symbols
                        ]
                        if len(really_new) > 0:
                            print(
                                f"Loading data for {len(really_new)} symbols: {really_new}"
                            )
                            rc_msg.append(
                                f"Loaded data for {len(really_new)} symbols: {really_new}"
                            )
                            self.minute_history = {
                                **self.minute_history,
                                **(
                                    market_data.get_historical_data_from_poylgon_for_symbols(
                                        self.data_api,
                                        really_new,
                                        self.start - timedelta(days=7),
                                        self.start + timedelta(days=1),
                                    )
                                ),
                            }
                            self.symbols += really_new
                            print(f"loaded data for {len(really_new)} stocks")

            for symbol in self.symbols:
                try:
                    for strategy in trading_data.strategies:

                        try:
                            minute_index = self.minute_history[symbol][
                                "close"
                            ].index.get_loc(self.now, method="nearest")
                        except Exception as e:
                            print(f"[Exception] {self.now} {symbol} {e}")
                            print(self.minute_history[symbol]["close"][-100:])
                            continue

                        price = self.minute_history[symbol]["close"][
                            minute_index
                        ]

                        if symbol not in trading_data.positions:
                            trading_data.positions[symbol] = 0

                        do, what = await strategy.run(
                            symbol,
                            True,
                            int(trading_data.positions[symbol]),
                            self.minute_history[symbol][: minute_index + 1],
                            self.now,
                            self.portfolio_value,
                            debug=False,  # type: ignore
                            backtesting=True,
                        )
                        if do:
                            if (
                                what["side"] == "buy"
                                and float(what["qty"]) > 0
                                or what["side"] == "sell"
                                and float(what["qty"]) < 0
                            ):
                                trading_data.positions[symbol] += int(
                                    float(what["qty"])
                                )
                                trading_data.buy_time[
                                    symbol
                                ] = self.now.replace(second=0, microsecond=0)
                            else:
                                trading_data.positions[symbol] -= int(
                                    float(what["qty"])
                                )

                            trading_data.last_used_strategy[symbol] = strategy

                            rc_msg.append(
                                f"[{self.now}][{strategy.name}] {what['side']} {what['qty']} of {symbol} @ {price}"
                            )
                            db_trade = NewTrade(
                                algo_run_id=strategy.algo_run.run_id,
                                symbol=symbol,
                                qty=int(float(what["qty"])),
                                operation=what["side"],
                                price=price,
                                indicators=trading_data.buy_indicators[symbol]
                                if what["side"] == "buy"
                                else trading_data.sell_indicators[symbol],
                            )

                            await db_trade.save(
                                config.db_conn_pool,
                                str(self.now.to_pydatetime()),
                                trading_data.stop_prices[symbol],
                                trading_data.target_prices[symbol],
                            )

                            if what["side"] == "buy":
                                await strategy.buy_callback(
                                    symbol, price, int(float(what["qty"]))
                                )
                                break
                            elif what["side"] == "sell":
                                await strategy.sell_callback(
                                    symbol, price, int(float(what["qty"]))
                                )
                                break
                except Exception as e:
                    print(f"[Exception] {self.now} {symbol} {e}")
                    traceback.print_exc()

            self.now += timedelta(minutes=1)

            return True, rc_msg
        else:
            return False, []

    async def liquidate(self):
        for symbol in trading_data.positions:
            if (
                trading_data.positions[symbol] != 0
                and trading_data.last_used_strategy[symbol].type
                == StrategyType.DAY_TRADE
            ):
                position = trading_data.positions[symbol]
                minute_index = self.minute_history[symbol][
                    "close"
                ].index.get_loc(self.now, method="nearest")
                price = self.minute_history[symbol]["close"][minute_index]
                tlog(f"[{self.end}]{symbol} liquidate {position} at {price}")
                db_trade = NewTrade(
                    algo_run_id=trading_data.last_used_strategy[symbol].algo_run.run_id,  # type: ignore
                    symbol=symbol,
                    qty=int(position) if int(position) > 0 else -int(position),
                    operation="sell" if position > 0 else "buy",
                    price=price,
                    indicators={"liquidate": 1},
                )
                await db_trade.save(
                    config.db_conn_pool, str(self.now.to_pydatetime())
                )



import asyncio
import time
from datetime import datetime, timedelta, date
from typing import List, Optional
from concurrent.futures import ThreadPoolExecutor
import alpaca_trade_api as tradeapi
import requests
from pytz import timezone

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog
from liualgotrader.models.ticker_data import StockOhlc
from liualgotrader.common.market_data import get_historical_daily_from_polygon_by_range

from .base import Scanner


class Momentum(Scanner):
    name = "momentum"

    def __init__(
        self,
        provider: str,
        recurrence: Optional[timedelta],
        target_strategy_name: Optional[str],
        data_api: tradeapi,
        max_share_price: float,
        min_share_price: float,
        min_last_dv: float,
        today_change_percent: float,
        min_volume: float,
        from_market_open: float,
        max_symbols: int = config.total_tickers,
        data_source: object = None,
    ):
        self.provider = provider
        self.max_share_price = max_share_price
        self.min_share_price = min_share_price
        self.min_last_dv = min_last_dv
        self.min_volume = min_volume
        self.today_change_percent = today_change_percent
        self.from_market_open = from_market_open
        self.max_symbols = max_symbols
        super().__init__(
            name=self.name,
            recurrence=recurrence,
            target_strategy_name=target_strategy_name,
            data_api=data_api,
            data_source=data_source,
        )

    @classmethod
    def __str__(cls) -> str:
        return cls.name

    async def _wait_time(self) -> None:
        if not config.bypass_market_schedule and config.market_open:
            nyc = timezone("America/New_York")
            since_market_open = datetime.today().astimezone(nyc) - config.market_open

            if since_market_open.seconds // 60 < self.from_market_open:
                tlog(f"market open, wait {self.from_market_open} minutes")
                while since_market_open.seconds // 60 < self.from_market_open:
                    await asyncio.sleep(1)
                    since_market_open = (
                        datetime.today().astimezone(nyc) - config.market_open
                    )

        tlog(f"Scanner {self.name} ready to run")

    def _get_trade_able_symbols(self) -> List[str]:
        assets = self.data_api.list_assets()
        tlog(f"loaded list of {len(assets)} trade-able assets from Alpaca")

        trade_able_symbols = [asset.symbol for asset in assets if asset.tradable]
        tlog(f"total number of trade-able symbols is {len(trade_able_symbols)}")
        return trade_able_symbols

    async def run_polygon(self) -> List[str]:
        tlog(f"{self.name}: run_polygon(): started")
        try:
            while True:
                tickers = self.data_api.polygon.all_tickers()
                tlog(f"loaded {len(tickers)} tickers from Polygon")
                if not len(tickers):
                    break
                trade_able_symbols = self._get_trade_able_symbols()

                unsorted = [
                    ticker
                    for ticker in tickers
                    if (
                        ticker.ticker in trade_able_symbols
                        and self.max_share_price
                        >= ticker.lastTrade["p"]
                        >= self.min_share_price
                        and ticker.prevDay["v"] * ticker.lastTrade["p"]
                        > self.min_last_dv
                        and ticker.todaysChangePerc >= self.today_change_percent
                        and (
                            ticker.day["v"] > self.min_volume
                            or config.bypass_market_schedule
                        )
                    )
                ]
                if len(unsorted) > 0:
                    ticker_by_volume = sorted(
                        unsorted,
                        key=lambda ticker: float(ticker.day["v"]),
                        reverse=True,
                    )
                    tlog(f"picked {len(ticker_by_volume)} symbols")
                    return [x.ticker for x in ticker_by_volume][: self.max_symbols]

                tlog("did not find gaping stock, retrying")
                await asyncio.sleep(30)
        except KeyboardInterrupt:
            tlog("KeyboardInterrupt")
            pass

        return []

    async def run_finnhub(self) -> List[str]:
        tlog(f"{self.name}: run_finnhub(): started")
        trade_able_symbols = self._get_trade_able_symbols()

        nyc = timezone("America/New_York")
        _from = datetime.today().astimezone(nyc) - timedelta(days=1)
        _to = datetime.now(nyc)
        symbols = []
        try:
            with requests.Session() as s:
                for symbol in trade_able_symbols:
                    try:
                        retry = True
                        while retry:
                            retry = False
                            url = (
                                f"{config.finnhub_base_url}/stock/candle?symbol={symbol}&"
                                f"resolution=D&from={_from.strftime('%s')}&to={_to.strftime('%s')}&"
                                f"token={config.finnhub_api_key}"
                            )

                            try:
                                r = s.get(url)
                            except ConnectionError:
                                retry = True
                                continue

                            if r.status_code == 200:
                                response = r.json()
                            if response["s"] != "no_data":
                                prev_prec = (
                                    100.0
                                    * (response["c"][1] - response["c"][0])
                                    / response["c"][0]
                                )
                                if (
                                    self.max_share_price
                                    > response["c"][1]
                                    > self.min_share_price
                                    and response["v"][0] * response["c"][0]
                                    > self.min_last_dv
                                    and prev_prec > self.today_change_percent
                                    and response["v"][1] > self.min_volume
                                ):
                                    symbols.append(symbol)
                                    tlog(f"collected {len(symbols)}/{self.max_symbols}")
                                    if len(symbols) == self.max_symbols:
                                        break

                            elif r.status_code == 429:
                                tlog(
                                    f"{trade_able_symbols.index(symbol)}/{len(trade_able_symbols)} API limit: ({r.text})"
                                )
                                time.sleep(30)
                                retry = True
                            else:
                                tlog(f"[ERROR] {r.status_code}, {r.text}")

                    except IndexError:
                        pass

        except KeyboardInterrupt:
            tlog("KeyboardInterrupt")
            pass

        tlog(f"loaded {len(symbols)} from Finnhub")
        return symbols

    async def add_stock_data_for_date(self, symbol: str, when: date) -> None:
        _minute_data = get_historical_daily_from_polygon_by_range(
            self.data_api, [symbol], when, when + timedelta(days=1)
        )

        await asyncio.sleep(0)
        if _minute_data[symbol].empty:
            daily_bar = StockOhlc(
                symbol=symbol,
                symbol_date=when,
                open=0.0,
                high=0.0,
                low=0.0,
                close=0.0,
                volume=0,
                indicators={},
            )
            await daily_bar.save()
        else:
            for index, row in _minute_data[symbol].iterrows():
                daily_bar = StockOhlc(
                    symbol=symbol,
                    symbol_date=index,
                    open=row["open"],
                    high=row["high"],
                    low=row["low"],
                    close=row["close"],
                    volume=int(row["volume"]),
                    indicators={},
                )
                await daily_bar.save()
                print(f"saved data for {symbol} @ {index}")

    async def fetch_symbol_details(
        self, symbol: str, back_time: datetime, session: requests.Session = None
    ):
        if not await StockOhlc.check_stock_date_exists(symbol, back_time):
            await self.add_stock_data_for_date(symbol, back_time)

        if not await StockOhlc.check_stock_date_exists(
            symbol, back_time - timedelta(days=1)
        ):
            await self.add_stock_data_for_date(symbol, back_time - timedelta(days=1))

    async def load_from_db(self, back_time: date) -> List[str]:
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                rows = await con.fetch(
                    """
                        SELECT
                            c2.symbol
                        FROM 
                            stock_ohlc as c1,
                            stock_ohlc as c2
                        WHERE
                            c1.symbol = c2.symbol AND 
                            c1.symbol_date = $1 AND
                            c2.symbol_date = $2 AND
                            c2.high < $3 AND
                            c2.low > $4 AND
                            c2.volume > $5 AND
                            c1.volume * c1.close > $6 AND
                            (c2.high / c1.close) > $7
                    """,
                    back_time - timedelta(days=1),
                    back_time,
                    self.max_share_price,
                    self.min_share_price,
                    self.min_volume,
                    self.min_last_dv,
                    1.0 + self.today_change_percent / 100.0,
                )

                if len(rows) > 0:
                    return [row[0] for row in rows]
                else:
                    return []

    async def run(self, back_time: datetime = None) -> List[str]:
        if not back_time:
            await self._wait_time()

            if self.provider == "polygon":
                return await self.run_polygon()
            elif self.provider == "finnhub":
                return await self.run_finnhub()
            else:
                raise Exception(
                    f"Invalid provider {self.provider} for scanner {self.name}"
                )
        else:
            rows = await self.load_from_db(back_time)

            if not len(rows):
                trade_able_symbols = self._get_trade_able_symbols()
                tlog(
                    f"{self.name} scanner => loading {len(trade_able_symbols)} symbols from Polygon and building cache. this may take a while."
                )
                tasks = [
                    asyncio.get_event_loop().create_task(
                        self.fetch_symbol_details(symbol, back_time, None)
                    )
                    for symbol in trade_able_symbols
                ]
                await asyncio.gather(*tasks)

                rows = await self.load_from_db(back_time)

            print(f"Scanner {self.name} -> back_time={back_time} picked {len(rows)}")
            return rows



import asyncio
import time
from concurrent.futures import ThreadPoolExecutor
from json.decoder import JSONDecodeError
from typing import Dict, List, Optional

import requests
from alpaca_trade_api.common import get_polygon_credentials
from alpaca_trade_api.polygon.entity import Ticker

from liualgotrader.common import config
from liualgotrader.common.decorators import timeit
from liualgotrader.common.tlog import tlog
from liualgotrader.miners.base import Miner
from liualgotrader.models.ticker_data import TickerData


class StockCluster(Miner):
    def __init__(self):
        self._num_workers = 20
        super().__init__(name="StockCluster")

    @property
    def num_workers(self):
        return self._num_workers

    @num_workers.setter
    def num_workers(self, new_num_workers: int):
        if new_num_workers <= 0 or new_num_workers > 100:
            raise ValueError("number of workers must be positive and less than 100")

    @timeit
    async def run(self) -> bool:
        tickers = []

        with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            with requests.Session() as session:
                count = self._get_count(session)
                loop = asyncio.get_event_loop()
                tasks = [
                    loop.run_in_executor(executor, self._fetch, *(session, page))
                    for page in range(1, count // 50 + 1)
                ]
                for response in await asyncio.gather(*tasks):
                    tickers += response
        tlog(f"loaded {len(tickers)} tickers")

        with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            with requests.Session() as session:
                loop = asyncio.get_event_loop()

                tasks = [
                    loop.run_in_executor(
                        executor,
                        self._fetch_symbol_details,  # type: ignore
                        *(session, ticker),
                    )
                    for ticker in tickers
                ]
                info = [
                    response
                    for response in await asyncio.gather(*tasks)
                    if response is not None
                ]
        tlog(f"loaded {len(info)} ticker details")
        await asyncio.gather(*[self._update_ticker_details(i) for i in info])

        return True

    def _get_count(self, session) -> int:
        url = "https://api.polygon.io/" + "v2" + "/reference/tickers"
        with session.get(
            url,
            params={
                "apiKey": get_polygon_credentials(config.prod_api_key_id),
                "market": "STOCKS",
                "page": 1,
                "active": "true",
                "perpage": 1,
            },
        ) as response:
            return response.json()["count"]

    def _fetch(self, session: requests.Session, page: int) -> List[Ticker]:
        url = "https://api.polygon.io/" + "v2" + "/reference/tickers"
        try:
            with session.get(
                url,
                params={
                    "apiKey": get_polygon_credentials(config.prod_api_key_id),
                    "market": "STOCKS",
                    "page": page,
                    "active": "true",
                    "perpage": 50,
                },
            ) as response:
                data = response.json()["tickers"]
                return [Ticker(x) for x in data]
        except requests.exceptions.ConnectionError as e:
            tlog(
                f"_fetch(): got HTTP exception {e}, for {page}, going to sleep, then retry"
            )
            time.sleep(30)
            return self._fetch(requests.Session(), page)

    async def _update_ticker_details(self, ticker_info: Dict) -> None:
        if ticker_info["active"] is False:
            return

        ticker_data = TickerData(
            name=ticker_info["name"],
            symbol=ticker_info["symbol"],
            description=ticker_info["description"],
            tags=ticker_info["tags"],
            similar_tickers=ticker_info["similar"],
            industry=ticker_info["industry"],
            sector=ticker_info["sector"],
            exchange=ticker_info["exchange"],
        )

        if await ticker_data.save(config.db_conn_pool) is False:
            tlog(f"going to wait 30 seconds and retry saving {ticker_info['name']}")
            await asyncio.sleep(30)
            return await self._update_ticker_details(ticker_info)

    def _fetch_symbol_details(
        self, session: requests.Session, ticker: Ticker
    ) -> Optional[Dict]:
        url = (
            "https://api.polygon.io/" + "v1" + f"/meta/symbols/{ticker.ticker}/company"
        )

        try:
            with session.get(
                url,
                params={"apiKey": get_polygon_credentials(config.prod_api_key_id)},
            ) as response:
                if response.status_code == 200:
                    try:
                        r = response.json()
                        if r["active"]:
                            return r
                    except JSONDecodeError:
                        tlog(f"JSONDecodeError for {ticker.ticker}")
                        raise Exception(response.text)

        except requests.exceptions.ConnectionError as e:
            tlog(
                f"_fetch_symbol_details(): got HTTP exception {e} for {ticker.ticker}, going to sleep, then retry"
            )
            time.sleep(30)
            return self._fetch_symbol_details(requests.Session(), ticker)

        return None



import asyncio
import importlib.util
import json
import multiprocessing as mp
import os
from datetime import datetime, timedelta
from typing import Dict, List

import alpaca_trade_api as tradeapi
from pytz import timezone
from pytz.tzinfo import DstTzInfo

from liualgotrader.common import config
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.tlog import tlog
from liualgotrader.scanners.base import Scanner
from liualgotrader.scanners.momentum import Momentum

scanner_tasks = []


async def scanner_runner(scanner: Scanner, queue: mp.Queue) -> None:
    try:
        while True:
            symbols = await scanner.run()

            if len(symbols):
                tlog(f"Scanner {scanner.name} picked {len(symbols)} symbols")
                queue.put(
                    json.dumps(
                        [
                            {
                                "symbol": symbol,
                                "target_strategy_name": scanner.target_strategy_name,
                            }
                            for symbol in symbols
                        ]
                    )
                )

            if scanner.recurrence:
                try:
                    await asyncio.sleep(scanner.recurrence.total_seconds())
                    tlog(f"scanner {scanner.name} re-running")
                except asyncio.CancelledError:
                    tlog(
                        f"scanner_runner({scanner.name}) cancelled during sleep, closing scanner task"
                    )
                    break
            else:
                break

    except asyncio.CancelledError:
        tlog(f"scanner_runner() cancelled, closing scanner task {scanner.name}")
    except Exception as e:
        tlog(
            f"[ERROR]Exception in scanner_runner({scanner.name}): exception of type {type(e).__name__} with args {e.args}"
        )
    finally:
        tlog(f"scanner_runner {scanner.name} completed")


async def scanners_runner(scanners_conf: Dict, queue: mp.Queue) -> None:
    data_api = tradeapi.REST(
        base_url=config.prod_base_url,
        key_id=config.prod_api_key_id,
        secret_key=config.prod_api_secret,
    )

    for scanner_name in scanners_conf:
        if scanner_name == "momentum":
            scanner_details = scanners_conf[scanner_name]
            try:
                recurrence = scanner_details.get("recurrence", None)
                target_strategy_name = scanner_details.get("target_strategy_name", None)
                scanner_object = Momentum(
                    provider=scanner_details["provider"],
                    data_api=data_api,
                    min_last_dv=scanner_details["min_last_dv"],
                    min_share_price=scanner_details["min_share_price"],
                    max_share_price=scanner_details["max_share_price"],
                    min_volume=scanner_details["min_volume"],
                    from_market_open=scanner_details["from_market_open"],
                    today_change_percent=scanner_details["min_gap"],
                    recurrence=timedelta(minutes=recurrence) if recurrence else None,
                    target_strategy_name=target_strategy_name,
                    max_symbols=scanner_details.get(
                        "max_symbols", config.total_tickers
                    ),
                )
                tlog(f"instantiated momentum scanner")
            except KeyError as e:
                tlog(
                    f"Error {e} in processing of scanner configuration {scanner_details}"
                )
                exit(0)
        else:
            tlog(f"custom scanner {scanner_name} selected")
            scanner_details = scanners_conf[scanner_name]
            try:
                spec = importlib.util.spec_from_file_location(
                    "module.name", scanner_details["filename"]
                )
                custom_scanner_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(custom_scanner_module)  # type: ignore
                class_name = scanner_name
                custom_scanner = getattr(custom_scanner_module, class_name)

                if not issubclass(custom_scanner, Scanner):
                    tlog(f"custom scanner must inherit from class {Scanner.__name__}")
                    exit(0)

                scanner_details.pop("filename")
                if "recurrence" not in scanner_details:
                    scanner_object = custom_scanner(
                        data_api=data_api,
                        **scanner_details,
                    )
                else:
                    recurrence = scanner_details.pop("recurrence")
                    scanner_object = custom_scanner(
                        data_api=data_api,
                        recurrence=timedelta(minutes=recurrence),
                        **scanner_details,
                    )

            except Exception as e:
                tlog(
                    f"[Error] scanners_runner.scanners_runner() for {scanner_name}:{e} "
                )

        scanner_tasks.append(asyncio.create_task(scanner_runner(scanner_object, queue)))

    try:
        await asyncio.gather(
            *scanner_tasks,
            return_exceptions=True,
        )

    except asyncio.CancelledError:
        tlog("scanners_runner.scanners_runner() cancelled, closing scanner tasks")

        for task in scanner_tasks:
            tlog(
                f"scanners_runner.scanners_runner()  requesting task {task.get_name()} to cancel"
            )
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                tlog("scanners_runner.scanners_runner()  task is cancelled now")

    finally:
        queue.close()
        tlog("scanners_runner.scanners_runner()  done.")


async def teardown_task(tz: DstTzInfo, tasks: List[asyncio.Task]) -> None:
    tlog("scanners_runner.teardown_task() starting")
    dt = datetime.today().astimezone(tz)
    to_market_close: timedelta
    try:
        to_market_close = (
            config.market_close - dt
            if config.market_close > dt
            else timedelta(hours=24) + (config.market_close - dt)
        )
        tlog(
            f"scanners_runner.teardown_task() task waiting for market close: {to_market_close}"
        )
    except Exception as e:
        tlog(
            f"scanners_runner.teardown_task() - exception of type {type(e).__name__} with args {e.args}"
        )
        return

    try:
        await asyncio.sleep(to_market_close.total_seconds() + 60 * 5)
        tlog("scanners_runner.teardown_task() closing tasks")

        for task in tasks:
            tlog(
                f"scanners_runner.teardown_task() requesting task {task.get_name()} to cancel"
            )
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                tlog("scanners_runner.teardown_task() task is cancelled now")

    except asyncio.CancelledError:
        tlog("scanners_runner.teardown_task() cancelled during sleep")
    finally:
        tlog("scanners_runner.teardown_task() done.")


async def async_main(scanners_conf: Dict, queue: mp.Queue) -> None:
    await create_db_connection(str(config.dsn))

    main_task = asyncio.create_task(
        scanners_runner(scanners_conf, queue),
        name="main_task",
    )

    tear_down = asyncio.create_task(
        teardown_task(
            timezone("America/New_York"),
            [main_task],
        )
    )

    await asyncio.gather(
        main_task,
        tear_down,
        return_exceptions=True,
    )


def main(
    conf_dict: Dict,
    market_open: datetime,
    market_close: datetime,
    scanner_queue: mp.Queue,
) -> None:
    tlog(f"*** scanners_runner.main() starting w pid {os.getpid()} ***")

    config.market_open = market_open
    config.market_close = market_close
    config.bypass_market_schedule = conf_dict.get("bypass_market_schedule", False)
    scanners_conf = conf_dict["scanners"]
    try:
        if not asyncio.get_event_loop().is_closed():
            asyncio.get_event_loop().close()
        asyncio.run(async_main(scanners_conf, scanner_queue))
    except KeyboardInterrupt:
        tlog("scanners_runner.main() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"scanners_runner.main() - exception of type {type(e).__name__} with args {e.args}"
        )

    tlog("*** scanners_runner.main() completed ***")



import os
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from asyncpg.pool import Pool

tradeplan_folder: str = (
    os.getenv("TRADEPLAN_DIR", ".") if len(os.getenv("TRADEPLAN_DIR", ".")) > 0 else "."
)
configuration_filename: str = "tradeplan.toml"
miner_configuration_filename: str = "miner.toml"


#
# Market Schedule
#
bypass_market_schedule: bool
market_open: datetime
market_close: datetime

# total number of tickers to follow
total_tickers = int(os.getenv("LIU_MAX_SYMBOLS", "100"))

#
# Shared data
#
build_label: str
filename: str
db_conn_pool: Pool
batch_id: str

#
# API keys
#
# Replace these with your API connection info from the dashboard
paper_base_url = os.getenv("ALPACA_PAPER_BASEURL", "https://paper-api.alpaca.markets")
paper_api_key_id = os.getenv("ALPACA_PAPER_API_KEY")
paper_api_secret = os.getenv("ALPACA_PAPER_API_SECRET")
finnhub_api_key = os.getenv("FINNHUB_API_KEY")
finnhub_base_url = os.getenv("FINNHUB_BASE_URL")
finnhub_websocket_limit = 50
#
# Execution details (env variable)
#
env: str = os.getenv("TRADE", "PAPER")
dsn: str = os.getenv("DSN", "")

prod_base_url = os.getenv("ALPACA_LIVE_BASEURL", "https://api.alpaca.markets")
prod_api_key_id = os.getenv("APCA_API_KEY_ID")
prod_api_secret = os.getenv("APCA_API_SECRET_KEY")


# Stop limit to default to
default_stop = 0.95

# How much of our portfolio to allocate to any one position
risk = 0.001
portfolio_value: Optional[float] = None

group_margin = 0.02

# trade times
market_liquidation_end_time_minutes: int = 15


#
# WS Data Channels
#
WS_DATA_CHANNELS: List[str]


# performance parameters
proc_factor: float = float(os.getenv("CPU_FACTOR", "2.0"))
num_consumers: int = int(os.getenv("NUM_CONSUMERS", "0"))

num_consumer_processes_ratio: int
# polygon parameters
polygon_seconds_timeout = 60


@dataclass
class polygon:
    MAX_DAYS_TO_LOAD: int = 7



import asyncio
import importlib.util
import json
import os
import sys
import traceback
from datetime import date, datetime, timedelta
from multiprocessing import Queue
from queue import Empty
from typing import Any, Dict, List

import alpaca_trade_api as tradeapi
import pandas as pd
import pygit2
from alpaca_trade_api.entity import Order
from alpaca_trade_api.rest import APIError
from pandas import DataFrame as df
from pytz import timezone
from pytz.tzinfo import DstTzInfo

from liualgotrader.common import config, market_data, trading_data
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.tlog import tlog
from liualgotrader.fincalcs.data_conditions import (QUOTE_SKIP_CONDITIONS,
                                                    TRADE_CONDITIONS)
from liualgotrader.models.new_trades import NewTrade
from liualgotrader.models.trending_tickers import TrendingTickers
from liualgotrader.strategies.base import Strategy, StrategyType

shortable: Dict = {}
symbol_data_error: Dict = {}


async def end_time(reason: str):
    for s in trading_data.strategies:
        tlog(f"updating end time for strategy {s.name}")
        await s.algo_run.update_end_time(
            pool=config.db_conn_pool, end_reason=reason
        )


async def is_shortable(trading_api: tradeapi, symbol: str) -> bool:

    asset = None
    while not asset:
        try:
            asset = trading_api.get_asset(symbol)
        except Exception as e:
            tlog(f"is_shortable({symbol}) got exception {e}, retrying...")
            await asyncio.sleep(10)

    return (
        False
        if asset.tradable is False
        or asset.shortable is False
        or asset.status == "inactive"
        or asset.easy_to_borrow is False
        else True
    )


async def liquidator(trading_api: tradeapi) -> None:
    tlog("liquidator() task starting")
    try:
        dt = datetime.today().astimezone(timezone("America/New_York"))
        to_market_close = (
            config.market_close - dt
            if config.market_close > dt
            else timedelta(hours=24) + (config.market_close - dt)
        ) - timedelta(minutes=15)
        await asyncio.sleep(to_market_close.total_seconds())

    except asyncio.CancelledError:
        tlog("liquidator() cancelled during sleep")
    except KeyboardInterrupt:
        tlog("liquidator() - Caught KeyboardInterrupt")

    tlog("liquidator() -> starting liqudation process")
    try:
        for symbol in trading_data.positions:
            tlog(f"liquidator() -> checking {symbol}")
            if (
                trading_data.positions[symbol] != 0
                and trading_data.last_used_strategy[symbol].type
                == StrategyType.DAY_TRADE
            ):
                await liquidate(
                    symbol, int(trading_data.positions[symbol]), trading_api
                )
    except asyncio.CancelledError:
        tlog("liquidator() cancelled")
    except KeyboardInterrupt:
        tlog("liquidator() - Caught KeyboardInterrupt")

    tlog("liquidator() task completed")


async def teardown_task(tz: DstTzInfo, task: asyncio.Task) -> None:
    tlog(f"consumer-teardown_task() - starting ")

    if not config.market_close:
        tlog(
            "we're probably in market schedule by-pass mode, exiting consumer-teardown_task()"
        )
        return

    to_market_close: timedelta
    try:
        dt = datetime.today().astimezone(tz)
        to_market_close = (
            config.market_close - dt
            if config.market_close > dt
            else timedelta(hours=24) + (config.market_close - dt)
        )
        tlog(
            f"consumer-teardown_task() - waiting for market close: {to_market_close}"
        )
    except Exception as e:
        tlog(
            f"consumer-teardown_task() - exception of type {type(e).__name__} with args {e.args}"
        )
        return

    try:
        await asyncio.sleep(to_market_close.total_seconds() + 60 * 5)

        tlog("consumer-teardown_task() starting")
        await end_time("market close")

        tlog("consumer-teardown_task(): requesting tasks to cancel")
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            tlog("consumer-teardown_task(): tasks are cancelled now")

    except asyncio.CancelledError:
        tlog("consumer-teardown_task() cancelled during sleep")
    except KeyboardInterrupt:
        tlog("consumer-teardown_task() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"consumer-teardown_task() - exception of type {type(e).__name__} with args {e.args}"
        )
        # asyncio.get_running_loop().stop()
    finally:
        tlog("consumer-teardown_task() task done.")


async def liquidate(
    symbol: str,
    symbol_position: int,
    trading_api: tradeapi,
) -> None:

    if symbol_position and symbol not in trading_data.open_orders:
        tlog(
            f"Trading over, trying to liquidate remaining position {symbol_position} in {symbol}"
        )
        try:
            if symbol_position < 0:
                o = trading_api.submit_order(
                    symbol=symbol,
                    qty=str(-symbol_position),
                    side="buy",
                    type="market",
                    time_in_force="day",
                )
                op = "buy"
                trading_data.buy_indicators[symbol] = {"liquidation": 1}

            else:
                o = trading_api.submit_order(
                    symbol=symbol,
                    qty=str(symbol_position),
                    side="sell",
                    type="market",
                    time_in_force="day",
                )
                op = "sell"
                trading_data.sell_indicators[symbol] = {"liquidation": 1}

            trading_data.open_orders[symbol] = (o, op)
            trading_data.open_order_strategy[
                symbol
            ] = trading_data.last_used_strategy[symbol]

        except Exception as e:
            tlog(f"failed to liquidate {symbol} w exception {e}")


async def should_cancel_order(order: Order, market_clock: datetime) -> bool:
    # Make sure the order's not too old
    submitted_at = order.submitted_at.astimezone(timezone("America/New_York"))
    order_lifetime = market_clock - submitted_at
    if market_clock > submitted_at and order_lifetime.seconds // 60 >= 1:
        return True

    return False


async def save(
    symbol: str,
    new_qty: int,
    last_op: str,
    price: float,
    indicators: Dict[Any, Any],
    now: str,
) -> None:
    db_trade = NewTrade(
        algo_run_id=trading_data.open_order_strategy[symbol].algo_run.run_id,
        symbol=symbol,
        qty=new_qty,
        operation=last_op,
        price=price,
        indicators=indicators,
    )

    await db_trade.save(
        config.db_conn_pool,
        str(now),
        trading_data.stop_prices[symbol],
        trading_data.target_prices[symbol],
    )


async def get_order(api: tradeapi, order_id: str) -> Order:
    return api.get_order(order_id)


async def update_partially_filled_order(
    strategy: Strategy, order: Order
) -> None:
    qty = int(order.filled_qty)
    new_qty = qty - abs(trading_data.partial_fills.get(order.symbol, 0))
    if order.side == "sell":
        qty = qty * -1

    trading_data.positions[order.symbol] = trading_data.positions.get(
        order.symbol, 0
    ) - trading_data.partial_fills.get(order.symbol, 0)
    trading_data.partial_fills[order.symbol] = qty
    trading_data.positions[order.symbol] += qty
    trading_data.open_orders[order.symbol] = (
        order,
        trading_data.open_orders.get(order.symbol)[1],  # type: ignore
    )

    try:
        indicators = {
            "buy": trading_data.buy_indicators.get(order.symbol, None),
            "sell": trading_data.sell_indicators.get(order.symbol, None),
        }
    except KeyError:
        indicators = {}

    await save(
        order.symbol,
        int(new_qty),
        order.side,
        float(order.filled_avg_price),
        indicators,
        order.updated_at,
    )

    if order.side == "buy":
        await strategy.buy_callback(
            order.symbol, float(order.filled_avg_price), int(new_qty)
        )
    else:
        await strategy.sell_callback(
            order.symbol, float(order.filled_avg_price), int(new_qty)
        )


async def update_filled_order(strategy: Strategy, order: Order) -> None:
    qty = int(order.filled_qty)
    new_qty = qty - abs(trading_data.partial_fills.get(order.symbol, 0))
    if order.side == "sell":
        qty = qty * -1

    trading_data.positions[order.symbol] = trading_data.positions.get(
        order.symbol, 0
    ) - trading_data.partial_fills.get(order.symbol, 0)
    trading_data.partial_fills[order.symbol] = 0
    trading_data.positions[order.symbol] += qty

    try:
        indicators = {
            "buy": trading_data.buy_indicators.get(order.symbol, None),
            "sell": trading_data.sell_indicators.get(order.symbol, None),
        }
    except KeyError:
        indicators = {}

    await save(
        order.symbol,
        int(new_qty),
        order.side,
        float(order.filled_avg_price),
        indicators,
        order.filled_at,
    )

    if order.side == "buy":
        trading_data.buy_indicators.pop(order.symbol, None)
        await strategy.buy_callback(
            order.symbol, float(order.filled_avg_price), int(new_qty)
        )
    else:
        trading_data.sell_indicators.pop(order.symbol, None)
        await strategy.sell_callback(
            order.symbol, float(order.filled_avg_price), int(new_qty)
        )

    trading_data.open_orders.pop(order.symbol, None)
    trading_data.open_order_strategy.pop(order.symbol, None)


async def handle_trade_update_for_order(data: Dict) -> bool:
    symbol = data["symbol"]

    last_order = trading_data.open_orders.get(symbol)[0]  # type: ignore
    if last_order is not None:
        event = data["event"]
        tlog(f"trade update for {symbol} data={data} with event {event}")

        if event == "partial_fill":
            await update_partially_filled_order(
                trading_data.open_order_strategy[symbol], Order(data["order"])
            )
        elif event == "fill":
            await update_filled_order(
                trading_data.open_order_strategy[symbol], Order(data["order"])
            )
        elif event in ("canceled", "rejected"):
            trading_data.partial_fills.pop(symbol, None)
            trading_data.open_orders.pop(symbol, None)
            trading_data.open_order_strategy.pop(symbol, None)

        return True
    else:
        tlog(
            f"[ERROR][{data['event']} trade update for {symbol} WITHOUT ORDER, should not arrive here"
        )
    return False


async def handle_trade_update_wo_order(data: Dict) -> bool:
    symbol = data["symbol"]
    event = data["event"]
    tlog(
        f"trade update without order for {symbol} data={data} with event {event}"
    )

    if event == "partial_fill":
        await update_partially_filled_order(
            trading_data.last_used_strategy[symbol], Order(data["order"])
        )
    elif event == "fill":
        await update_filled_order(
            trading_data.last_used_strategy[symbol], Order(data["order"])
        )
    elif event in ("canceled", "rejected"):
        trading_data.partial_fills.pop(symbol, None)

    return True


async def handle_trade_update(data: Dict) -> bool:
    symbol = data["symbol"]
    if trading_data.open_orders.get(symbol):
        return await handle_trade_update_for_order(data)
    else:
        return await handle_trade_update_wo_order(data)


async def handle_data_queue_msg(
    data: Dict, trading_api: tradeapi, data_api: tradeapi
) -> bool:
    global shortable
    global symbol_data_error

    symbol = data["symbol"]
    if symbol not in market_data.minute_history:
        _df = data_api.polygon.historic_agg_v2(
            symbol,
            1,
            "minute",
            _from=str(date.today() - timedelta(days=10)),
            to=str(date.today() + timedelta(days=1)),
        ).df
        _df["vwap"] = 0.0
        _df["average"] = 0.0
        market_data.minute_history[symbol] = _df
        tlog(
            f"consumer task loaded {len(market_data.minute_history[symbol].index)} 1-min candles for {symbol}"
        )
        shortable[symbol] = True  # await is_shortable(data_api, symbol)
    elif not shortable.get(symbol):
        shortable[symbol] = True  # await is_shortable(data_api, symbol)

    if data["EV"] == "T":
        if "conditions" in data and any(
            item in data["conditions"] for item in TRADE_CONDITIONS
        ):
            # tlog(f"trade={data}")
            return True
        return True
    elif data["EV"] == "Q":
        if "askprice" not in data or "bidprice" not in data:
            return True
        if "condition" in data and any(
            item == data["condition"] for item in QUOTE_SKIP_CONDITIONS
        ):
            return True

        # tlog(f"quote={data}")
        prev_ask = trading_data.voi_ask.get(symbol, None)
        prev_bid = trading_data.voi_bid.get(symbol, None)
        trading_data.voi_ask[symbol] = (
            data["askprice"],
            data["asksize"],
            data["timestamp"],
        )
        trading_data.voi_bid[symbol] = (
            data["bidprice"],
            data["bidsize"],
            data["timestamp"],
        )

        bid_delta_volume = (
            0
            if not prev_bid or data["bidprice"] < prev_bid[0]
            else 100 * data["bidsize"]
            if data["bidprice"] > prev_bid[0]
            else 100 * (data["bidsize"] - prev_bid[1])
        )
        ask_delta_volume = (
            0
            if not prev_ask or data["askprice"] > prev_ask[0]
            else 100 * data["asksize"]
            if data["askprice"] < prev_ask[0]
            else 100 * (data["asksize"] - prev_ask[1])
        )
        voi_stack = trading_data.voi.get(symbol, None)
        if not voi_stack:
            voi_stack = [0.0]
        elif len(voi_stack) == 10:
            voi_stack[0:9] = voi_stack[1:10]
            voi_stack.pop()

        k = 2.0 / (100 + 1)
        voi_stack.append(
            round(
                voi_stack[-1] * (1.0 - k)
                + k * (bid_delta_volume - ask_delta_volume),
                2,
            )
        )
        trading_data.voi[symbol] = voi_stack
        # tlog(f"{symbol} voi:{trading_data.voi[symbol]}")

    elif data["EV"] in ("A", "AM"):
        original_ts = ts = pd.Timestamp(
            data["start"], tz="America/New_York", unit="ms"
        )
        ts = ts.replace(second=0, microsecond=0)

        try:
            current = market_data.minute_history[symbol].loc[ts]
        except KeyError:
            current = None

        if current is None:
            new_data = [
                data["open"],
                data["high"],
                data["low"],
                data["close"],
                data["volume"],
                data["vwap"],
                data["average"],
            ]
        else:
            new_data = [
                current.open,
                max(data["high"], current.high),
                min(data["low"], current.low),
                data["close"],
                current.volume + data["volume"],
                data["vwap"],
                data["average"],
            ]
        market_data.minute_history[symbol].loc[ts] = new_data
        market_data.volume_today[symbol] = data["totalvolume"]

        if data["EV"] == "A":
            if (time_diff := datetime.now(tz=timezone("America/New_York")) - original_ts) > timedelta(seconds=10):  # type: ignore
                tlog(f"A$ {symbol} too out of sync w {time_diff}")
                return False
            elif (
                curr_min := datetime.now(
                    tz=timezone("America/New_York")
                ).replace(second=0, microsecond=0)
            ) > ts:
                return True
        elif data["EV"] == "AM":
            return True

        # Next, check for existing orders for the stock
        existing_order = trading_data.open_orders.get(symbol)

        if existing_order is not None:
            existing_order = existing_order[0]
            try:
                if await should_cancel_order(existing_order, original_ts):
                    inflight_order = await get_order(
                        trading_api, existing_order.id  # type: ignore
                    )
                    if inflight_order and inflight_order.status == "filled":
                        tlog(
                            f"order_id {existing_order.id} for {symbol} already filled {inflight_order}"  # type: ignore
                        )
                        await update_filled_order(
                            trading_data.open_order_strategy[symbol],
                            inflight_order,
                        )
                    elif (
                        inflight_order
                        and inflight_order.status == "partially_filled"
                    ):
                        tlog(
                            f"order_id {existing_order.id} for {symbol} already partially_filled {inflight_order}"  # type: ignore
                        )
                        await update_partially_filled_order(
                            trading_data.open_order_strategy[symbol],
                            inflight_order,
                        )
                    else:
                        # Cancel it so we can try again for a fill
                        tlog(
                            f"Cancel order id {existing_order.id} for {symbol} ts={original_ts} submission_ts={existing_order.submitted_at.astimezone(timezone('America/New_York'))}"  # type: ignore
                        )
                        trading_api.cancel_order(existing_order.id)  # type: ignore
                        trading_data.open_orders.pop(symbol, None)

                return True
            except AttributeError:
                tlog(f"Attribute Error in symbol {symbol} w/ {existing_order}")

        # do we have a position?
        symbol_position = trading_data.positions.get(symbol, 0)

        # do we need to liquidate for the day?
        until_market_close = config.market_close - ts
        if (
            until_market_close.seconds // 60
            <= config.market_liquidation_end_time_minutes
            and symbol_position != 0
            and trading_data.last_used_strategy[symbol].type
            == StrategyType.DAY_TRADE
        ):
            await liquidate(symbol, int(symbol_position), trading_api)
        else:
            # run strategies
            for s in trading_data.strategies:
                if (
                    "symbol_strategy" in data
                    and data["symbol_strategy"]
                    and s.name != data["symbol_strategy"]
                ):
                    continue

                try:
                    do, what = await s.run(
                        symbol,
                        shortable[symbol],
                        int(symbol_position),
                        market_data.minute_history[symbol],
                        ts,
                        trading_api=trading_api,
                        portfolio_value=config.portfolio_value,
                    )
                except Exception as e:
                    # exc_info = sys.exc_info()
                    # lines = traceback.format_exception(*exc_info)
                    # for line in lines:
                    #    tlog(f"{line}")
                    # del exc_info

                    symbol_data_error[symbol] = (
                        0
                        if symbol not in symbol_data_error
                        else symbol_data_error[symbol] + 1
                    )
                    tlog(
                        f"[EXCEPTION] strategy {s.name} for symbol {symbol} -> {e} [{symbol_data_error[symbol]}]"
                    )
                    if symbol_data_error[symbol] < 5:
                        tlog(f"attempting reload of data for symbol {symbol}")

                        _df = data_api.polygon.historic_agg_v2(
                            symbol,
                            1,
                            "minute",
                            _from=str(date.today() - timedelta(days=10)),
                            to=str(date.today() + timedelta(days=1)),
                        ).df
                        _df["vwap"] = 0.0
                        _df["average"] = 0.0
                        market_data.minute_history[symbol] = _df
                        tlog(
                            f"consumer task re-loaded {len(market_data.minute_history[symbol].index)} 1-min candles for {symbol}"
                        )
                    continue

                if do:
                    try:
                        if what["type"] == "limit":
                            o = trading_api.submit_order(
                                symbol=symbol,
                                qty=what["qty"],
                                side=what["side"],
                                type="limit",
                                time_in_force="day",
                                limit_price=what["limit_price"],
                            )
                        else:
                            o = trading_api.submit_order(
                                symbol=symbol,
                                qty=what["qty"],
                                side=what["side"],
                                type=what["type"],
                                time_in_force="day",
                            )

                        trading_data.open_orders[symbol] = (o, what["side"])
                        trading_data.open_order_strategy[symbol] = s

                        tlog(
                            f"executed strategy {s.name} on {symbol} w data {market_data.minute_history[symbol][-10:]}"
                        )
                        trading_data.last_used_strategy[symbol] = s
                        if what["side"] == "buy":
                            trading_data.buy_time[symbol] = datetime.now(
                                tz=timezone("America/New_York")
                            ).replace(second=0, microsecond=0)
                            break
                    except APIError as e:
                        tlog(
                            f"Exception APIError with {e} from {what}, checking if order filled"
                        )

    return True


async def queue_consumer(
    queue: Queue,
    trading_api: tradeapi,
    data_api: tradeapi,
) -> None:
    tlog("queue_consumer() starting")

    try:
        while True:
            try:
                raw_data = queue.get(timeout=2)
                data = json.loads(raw_data)

                if data["EV"] == "trade_update":
                    tlog(f"received trade_update: {data}")
                    await handle_trade_update(data)
                else:
                    if not await handle_data_queue_msg(
                        data, trading_api, data_api
                    ):
                        while not queue.empty():
                            _ = queue.get()
                        tlog("cleaned queue")

            except Empty:
                await asyncio.sleep(0)
                continue
            except Exception as e:
                tlog(
                    f"Exception in queue_consumer(): exception of type {type(e).__name__} with args {e.args} inside loop"
                )
                exc_info = sys.exc_info()
                lines = traceback.format_exception(*exc_info)
                for line in lines:
                    tlog(f"error: {line}")
                traceback.print_exception(*exc_info)
                del exc_info

    except asyncio.CancelledError:
        tlog("queue_consumer() cancelled ")
    except Exception as e:
        tlog(
            f"Exception in queue_consumer(): exception of type {type(e).__name__} with args {e.args}"
        )
        exc_info = sys.exc_info()
        lines = traceback.format_exception(*exc_info)
        for line in lines:
            tlog(f"error: {line}")
        traceback.print_exception(*exc_info)
        del exc_info
    finally:
        tlog("queue_consumer() task done.")


def get_trading_windows(tz, api):
    """Get start and end time for trading"""

    today = datetime.today().astimezone(tz)
    today_str = datetime.today().astimezone(tz).strftime("%Y-%m-%d")

    calendar = api.get_calendar(start=today_str, end=today_str)[0]

    tlog(f"next open date {calendar.date.date()}")

    if today.date() < calendar.date.date():
        tlog(f"which is not today {today}")
        return None, None
    market_open = today.replace(
        hour=calendar.open.hour,
        minute=calendar.open.minute,
        second=0,
        microsecond=0,
    )
    market_close = today.replace(
        hour=calendar.close.hour,
        minute=calendar.close.minute,
        second=0,
        microsecond=0,
    )
    return market_open, market_close


async def consumer_async_main(
    queue: Queue,
    symbols: List[str],
    unique_id: str,
    strategies_conf: Dict,
):
    await create_db_connection(str(config.dsn))

    if symbols:
        try:
            trending_db = TrendingTickers(unique_id)
            await trending_db.save(symbols)
        except Exception as e:
            tlog(
                f"Exception in consumer_async_main() while storing symbols to DB:{type(e).__name__} with args {e.args}"
            )
            exc_info = sys.exc_info()
            lines = traceback.format_exception(*exc_info)
            for line in lines:
                tlog(f"error: {line}")
            traceback.print_exception(*exc_info)
            del exc_info

    base_url = (
        config.prod_base_url if config.env == "PROD" else config.paper_base_url
    )
    api_key_id = (
        config.prod_api_key_id
        if config.env == "PROD"
        else config.paper_api_key_id
    )
    api_secret = (
        config.prod_api_secret
        if config.env == "PROD"
        else config.paper_api_secret
    )
    trading_api = tradeapi.REST(
        base_url=base_url, key_id=api_key_id, secret_key=api_secret
    )
    data_api = tradeapi.REST(
        base_url=config.prod_base_url,
        key_id=config.prod_api_key_id,
        secret_key=config.prod_api_secret,
    )
    nyc = timezone("America/New_York")
    config.market_open, config.market_close = get_trading_windows(
        nyc, trading_api
    )
    strategy_types = []
    for strategy_name in strategies_conf:
        strategy_details = strategies_conf[strategy_name]
        tlog(f"custom strategy {strategy_name} selected")

        try:
            spec = importlib.util.spec_from_file_location(
                "module.name", strategy_details["filename"]
            )
            custom_strategy_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(custom_strategy_module)  # type: ignore
            class_name = strategy_name

            custom_strategy = getattr(custom_strategy_module, class_name)

            if not issubclass(custom_strategy, Strategy):
                tlog(
                    f"custom strartegy must inherit from class {Strategy.__name__}"
                )
                exit(0)
            strategy_details.pop("filename", None)
            strategy_types += [(custom_strategy, strategy_details)]

        except FileNotFoundError as e:
            tlog(f"[Error] file not found `{strategy_details['filename']}`")
            exit(0)
        except Exception as e:
            tlog(
                f"[Error]exception of type {type(e).__name__} with args {e.args}"
            )
            exit(0)

    loaded = 0
    for strategy_tuple in strategy_types:
        strategy_type = strategy_tuple[0]
        strategy_details = strategy_tuple[1]
        tlog(f"initializing {strategy_type.name}")
        s = strategy_type(batch_id=unique_id, **strategy_details)
        await s.create()

        trading_data.strategies.append(s)
        if symbols:
            loaded += await load_current_positions(
                trading_api=trading_api,
                symbols=symbols,
                strategy=s,
                env=config.env,
            )

    if symbols and loaded != len(symbols):
        tlog(
            f"[ERROR] Consumer process loaded only {loaded} out of {len(symbols)} open positions. HINT: make sure that your tradeplan.toml file includes all strategues in previous trading session."
        )

    queue_consumer_task = asyncio.create_task(
        queue_consumer(queue, trading_api, data_api)
    )

    liquidate_task = asyncio.create_task(liquidator(trading_api))

    tear_down = asyncio.create_task(
        teardown_task(timezone("America/New_York"), queue_consumer_task)
    )
    await asyncio.gather(
        tear_down,
        liquidate_task,
        queue_consumer_task,
        return_exceptions=True,
    )

    tlog("consumer_async_main() completed")


async def load_current_positions(
    trading_api: tradeapi, symbols: List[str], strategy: Strategy, env: str
) -> int:
    loaded = 0
    for symbol in symbols:
        try:
            position = trading_api.get_position(symbol)
        except Exception as e:
            tlog(f"failed to load open position for {symbol} w/ {e}")
            continue

        if position:

            try:
                (
                    prev_run_id,
                    price,
                    stop_price,
                    target_price,
                    indicators,
                    timestamp,
                ) = await NewTrade.load_latest(
                    config.db_conn_pool,
                    symbol=symbol,
                    strategy_name=strategy.name,
                    env=env,
                )

                if prev_run_id is None:
                    continue

                tlog(
                    f"loading current position for {symbol} for strategy {strategy.name}"
                )

                trading_data.positions[symbol] = int(position.qty)
                trading_data.stop_prices[symbol] = stop_price
                trading_data.target_prices[symbol] = target_price
                trading_data.latest_cost_basis[
                    symbol
                ] = trading_data.latest_scalp_basis[symbol] = price
                trading_data.open_order_strategy[symbol] = strategy
                trading_data.last_used_strategy[symbol] = strategy
                trading_data.buy_time[symbol] = timestamp.astimezone(
                    tz=timezone("America/New_York")
                )

                await NewTrade.rename_algo_run_id(
                    strategy.algo_run.run_id, prev_run_id, symbol
                )
                tlog(
                    f"moved {symbol} from {prev_run_id} to {strategy.algo_run.run_id}"
                )

                loaded += 1

            except ValueError:
                pass
            except Exception as e:
                tlog(
                    f"load_current_positions() for {symbol} could not load latest trade from db due to exception of type {type(e).__name__} with args {e.args}"
                )

    return loaded


def consumer_main(
    queue: Queue,
    symbols: List[str],
    minute_history: Dict[str, df],
    unique_id: str,
    conf: Dict,
) -> None:
    tlog(f"*** consumer_main() starting w pid {os.getpid()} ***")

    try:
        config.build_label = pygit2.Repository("../").describe(
            describe_strategy=pygit2.GIT_DESCRIBE_TAGS
        )
    except pygit2.GitError:
        import liualgotrader

        config.build_label = liualgotrader.__version__ if hasattr(liualgotrader, "__version__") else ""  # type: ignore

    config.bypass_market_schedule = conf.get("bypass_market_schedule", False)
    config.portfolio_value = conf.get("portfolio_value", None)
    if "risk" in conf:
        config.risk = conf["risk"]
    if "market_liquidation_end_time_minutes" in conf:
        config.market_liquidation_end_time_minutes = conf[
            "market_liquidation_end_time_minutes"
        ]

    market_data.minute_history = minute_history
    try:
        if not asyncio.get_event_loop().is_closed():
            asyncio.get_event_loop().close()
        asyncio.run(
            consumer_async_main(queue, symbols, unique_id, conf["strategies"])
        )
    except KeyboardInterrupt:
        tlog("consumer_main() - Caught KeyboardInterrupt")

    tlog("*** consumer_main() completed ***")



from abc import ABCMeta, abstractmethod


class Miner(metaclass=ABCMeta):
    def __init__(
        self,
        name: str,
    ):
        self._name = name

    @property
    def name(self):
        return self._name

    @abstractmethod
    async def run(self) -> bool:
        pass



from typing import Dict

import asyncpg
import pandas as pd

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog


async def create_db_connection(dsn: str = None) -> None:
    config.db_conn_pool = await asyncpg.create_pool(
        dsn=dsn if dsn else config.dsn,
        min_size=2,
        max_size=10,
        command_timeout=30,
    )
    tlog("db connection pool initialized")


async def fetch_as_dataframe(query: str, *args) -> pd.DataFrame:
    try:
        if config.db_conn_pool:
            pass
    except AttributeError:
        await create_db_connection()

    async with config.db_conn_pool.acquire() as con:
        stmt = await con.prepare(query)
        columns = [a.name for a in stmt.get_attributes()]
        data = await stmt.fetch(*args)
        return pd.DataFrame(data=data, columns=columns)



"""Global data shared during trading"""
from asyncio import Queue
from datetime import datetime
from typing import Dict, List, Optional, Tuple

from alpaca_trade_api.entity import Order

from liualgotrader.models.ticker_snapshot import TickerSnapshot
from liualgotrader.strategies.base import Strategy

strategies: List[Strategy] = []
open_orders: Dict[str, Tuple[Order, str]] = {}
open_order_strategy: Dict[str, Strategy] = {}
last_used_strategy: Dict[str, Strategy] = {}
latest_cost_basis: Dict[str, float] = {}
latest_scalp_basis: Dict[str, float] = {}
sell_indicators: Dict[str, Dict] = {}
buy_indicators: Dict[str, Dict] = {}
positions: Dict[str, float] = {}
target_prices: Dict[str, float] = {}
stop_prices: Dict[str, float] = {}
partial_fills: Dict[str, float] = {}
symbol_resistance: Dict[str, float] = {}
voi: Dict[str, List[float]] = {}
voi_ask: Dict[str, Tuple[float, float, datetime]] = {}
voi_bid: Dict[str, Tuple[float, float, datetime]] = {}


industry_trend: Dict[str, float] = {}
sector_trend: Dict[str, float] = {}
snapshot: Dict[str, TickerSnapshot] = {}

cool_down: Dict[str, Optional[datetime]] = {}

queues: Dict[str, Queue] = {}
down_cross: Dict[str, datetime] = {}

buy_time: Dict[str, datetime] = {}



"""
Get Market data from Polygon and pump to consumers
"""
import asyncio
import json
import os
import random
import sys
import traceback
from datetime import datetime, timedelta
from multiprocessing import Queue
from queue import Empty, Full
from typing import Dict, List

import alpaca_trade_api as tradeapi
from alpaca_trade_api.stream2 import StreamConn, polygon
from pytz import timezone
from pytz.tzinfo import DstTzInfo

from liualgotrader.common import config
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.tlog import tlog
from liualgotrader.models.trending_tickers import TrendingTickers

last_msg_tstamp: datetime = datetime.now()
symbols: List[str]
data_channels: List = []
queue_id_hash: Dict[str, int]
symbol_strategy: Dict = {}


async def scanner_input(
    scanner_queue: Queue,
    data_ws: StreamConn,
    num_consumer_processes: int,
) -> None:
    tlog("scanner_input() task starting ")
    global data_channels
    global queue_id_hash
    global symbol_strategy
    global symbols

    while True:
        try:
            symbols_details = scanner_queue.get(timeout=1)
            if symbols_details:
                symbols_details = json.loads(symbols_details)
                new_symbols: List = []
                new_channels: List = []
                for symbol_details in symbols_details:
                    if symbol_details["symbol"] not in symbols:
                        new_symbols.append(symbol_details["symbol"])
                        symbol_strategy[
                            symbol_details["symbol"]
                        ] = symbol_details["target_strategy_name"]
                        new_channels += [
                            f"{OP}.{symbol_details['symbol']}"
                            for OP in config.WS_DATA_CHANNELS
                        ]
                        consumer_queue_index = random.SystemRandom().randint(
                            0, num_consumer_processes - 1
                        )
                        queue_id_hash[
                            symbol_details["symbol"]
                        ] = consumer_queue_index

                if len(new_symbols):
                    symbols += new_symbols
                    data_channels += new_channels

                    retry = 5
                    while retry > 0:
                        try:
                            await data_ws.subscribe(new_channels)
                            break
                        except Exception as e:
                            tlog(f"[EXCEPTION] {e} below, retrying {retry}")
                            exc_info = sys.exc_info()
                            lines = traceback.format_exception(*exc_info)
                            for line in lines:
                                tlog(f"error: {line}")
                            await asyncio.sleep(1)
                            retry -= 1

                    trending_db = TrendingTickers(config.batch_id)
                    await trending_db.save(new_symbols)

                    tlog(f"added {len(new_symbols)}:{new_symbols}")
                    await asyncio.sleep(1)

        except Empty:
            await asyncio.sleep(30)
        except asyncio.CancelledError:
            tlog("scanner_input() task task cancelled ")
            break
        except Exception as e:
            tlog(
                f"Exception in scanner_input(): exception of type {type(e).__name__} with args {e.args}"
            )
            exc_info = sys.exc_info()
            lines = traceback.format_exception(*exc_info)
            for line in lines:
                tlog(f"error: {line}")
            traceback.print_exception(*exc_info)
            del exc_info

    tlog("scanner_input() task completed")


async def trade_run(
    ws: StreamConn,
    queues: List[Queue],
) -> None:

    tlog("trade_run() starting using Alpaca trading  ")

    @ws.on(r"trade_update")
    async def handle_trade_update(conn, channel, data):
        global queue_id_hash
        try:
            # tlog(f"producer TRADE UPDATE event: {data.__dict__}")
            symbol = data.__dict__["_raw"]["order"]["symbol"]
            if qid := queue_id_hash.get(symbol, None):
                data.__dict__["_raw"]["EV"] = "trade_update"
                data.__dict__["_raw"]["symbol"] = symbol
                queues[qid].put(json.dumps(data.__dict__["_raw"]))

        except Exception as e:
            tlog(
                f"[ERROR]Exception in handle_trade_update(): exception of type {type(e).__name__} with args {e.args}"
            )
            traceback.print_exc()

    await ws.subscribe(["trade_updates"])
    tlog("trade_run() completed")


async def run(
    data_ws: StreamConn,
    queues: List[Queue],
) -> None:
    global data_channels
    global queue_id_hash
    for symbol in symbols:
        symbol_channels = [f"{OP}.{symbol}" for OP in config.WS_DATA_CHANNELS]
        data_channels += symbol_channels

    tlog(f"Watching {len(symbols)} symbols from Polygon.io")

    @data_ws.on(r"T$")
    async def handle_trade_event(conn, channel, data):
        global last_msg_tstamp
        last_msg_tstamp = datetime.now()

        queue_id: int = -1
        try:
            if (time_diff := datetime.now(tz=timezone("America/New_York")) - data.timestamp) > timedelta(seconds=10):  # type: ignore
                return
            elif (event_symbol := data.__dict__["_raw"]["symbol"]) in queue_id_hash:  # type: ignore
                data.__dict__["_raw"]["EV"] = "T"
                queue_id = queue_id_hash[event_symbol]
                queues[queue_id].put(
                    json.dumps(data.__dict__["_raw"]), timeout=1
                )
        except Full as f:
            tlog(
                f"[EXCEPTION] handle_trade_event():queue {queue_id} is FULL:{f}, sleep for 2 seconds and re-try."
            )
            await asyncio.sleep(2)
            await handle_trade_event(conn, channel, data)
        except Exception as e:
            tlog(
                f"Exception in handle_trade_event(): exception of type {type(e).__name__} with args {e.args}"
            )
            print(queue_id, len(queues))
            traceback.print_exc()

    @data_ws.on(r"Q$")
    async def handle_quote_event(conn, channel, data):
        global last_msg_tstamp
        last_msg_tstamp = datetime.now()

        queue_id: int = -1
        try:
            if (time_diff := datetime.now(tz=timezone("America/New_York")) - data.timestamp) > timedelta(seconds=10):  # type: ignore
                return
            elif (event_symbol := data.__dict__["_raw"]["symbol"]) in queue_id_hash:  # type: ignore
                data.__dict__["_raw"]["EV"] = "Q"
                queue_id = queue_id_hash[event_symbol]
                queues[queue_id].put(
                    json.dumps(data.__dict__["_raw"]), timeout=1
                )

        except Full as f:
            tlog(
                f"[EXCEPTION] handle_quote_event():queue {queue_id} is FULL:{f}, sleeping for 2 seconds and re-trying."
            )
            await asyncio.sleep(2)
            await handle_quote_event(conn, channel, data)

        except Exception as e:
            tlog(
                f"Exception in handle_quote_event(): exception of type {type(e).__name__} with args {e.args}"
            )
            print(queue_id, len(queues))
            traceback.print_exc()

    @data_ws.on(r"A$")
    async def handle_second_bar(conn, channel, data):
        global last_msg_tstamp
        global symbol_strategy
        last_msg_tstamp = datetime.now()

        queue_id: int = -1
        try:
            if (time_diff := datetime.now(tz=timezone("America/New_York")) - data.start) > timedelta(seconds=8):  # type: ignore
                # tlog(f"A$ {data.symbol}: data out of sync {time_diff}")
                pass
            elif (event_symbol := data.__dict__["_raw"]["symbol"]) in queue_id_hash:  # type: ignore
                data.__dict__["_raw"]["EV"] = "A"
                if (
                    event_symbol in symbol_strategy
                    and symbol_strategy[event_symbol]
                ):
                    data.__dict__["_raw"]["symbol_strategy"] = symbol_strategy[
                        event_symbol
                    ]
                queue_id = queue_id_hash[event_symbol]
                queues[queue_id].put(
                    json.dumps(data.__dict__["_raw"]), timeout=1
                )
        except Full as f:
            tlog(
                f"[EXCEPTION] handle_second_bar(): queue {queue_id} is FULL:{f}, sleeping for 2 seconds and re-trying."
            )
            await asyncio.sleep(2)
            await handle_second_bar(conn, channel, data)
        except Exception as e:
            tlog(
                f"Exception in handle_second_bar(): exception of type {type(e).__name__} with args {e.args}"
            )
            print(queue_id, len(queues))
            traceback.print_exc()

    @data_ws.on(r"AM$")
    async def handle_minute_bar(conn, channel, data):
        global last_msg_tstamp
        global symbol_strategy
        last_msg_tstamp = datetime.now()

        queue_id: int = -1
        try:
            if (event_symbol := data.__dict__["_raw"]["symbol"]) in queue_id_hash:  # type: ignore
                data.__dict__["_raw"]["EV"] = "AM"
                if (
                    event_symbol in symbol_strategy
                    and symbol_strategy[event_symbol]
                ):
                    data.__dict__["_raw"]["symbol_strategy"] = symbol_strategy[
                        event_symbol
                    ]
                queue_id = queue_id_hash[event_symbol]
                queues[queue_id].put(
                    json.dumps(data.__dict__["_raw"]), timeout=1
                )

        except Full as f:
            tlog(
                f"[EXCEPTION] handle_minute_bar(): queue {queue_id} is FULL:{f}, sleeping for 2 seconds and re-trying."
            )
            await asyncio.sleep(2)
            await handle_minute_bar(conn, channel, data)

        except Exception as e:
            tlog(
                f"Exception in handle_minute_bar(): exception of type {type(e).__name__} with args {e.args}"
            )
            traceback.print_exc()

    global last_msg_tstamp
    try:
        await data_ws.subscribe(data_channels)

        while True:
            # print(f"tick! {datetime.now() - last_msg_tstamp}")
            if (datetime.now() - last_msg_tstamp) > timedelta(
                seconds=config.polygon_seconds_timeout
            ):
                tlog(
                    f"no data activity since {last_msg_tstamp} attempting reconnect"
                )
                await data_ws.close(False)
                data_ws.data_ws = polygon.StreamConn(config.prod_api_key_id)
                await data_ws.data_ws.connect()
                data_ws.register(r"AM$", handle_minute_bar)
                data_ws.register(r"A$", handle_second_bar)
                data_ws.register(r"Q$", handle_quote_event)
                data_ws.register(r"T$", handle_trade_event)
                await data_ws.subscribe(data_channels)
                tlog(
                    f"Polygon.io reconnected for {len(data_channels)} channels"
                )
                last_msg_tstamp = datetime.now()
            await asyncio.sleep(config.polygon_seconds_timeout / 2)
    except asyncio.CancelledError:
        tlog("main Polygon.io consumer task cancelled ")
    except Exception as e:
        tlog(
            f"Exception in Polygon.io consumer task: exception of type {type(e).__name__} with args {e.args}"
        )
        traceback.print_exc()
    finally:
        for q in queues:
            q.close()
        tlog("" "main Polygon producer task completed ")


async def teardown_task(
    tz: DstTzInfo, ws: List[StreamConn], tasks: List[asyncio.Task]
) -> None:
    tlog("poylgon_producer teardown_task() starting")
    if not config.market_close:
        tlog(
            "we're probably in market schedule by-pass mode, exiting poylgon_producer tear-down task"
        )
        return

    dt = datetime.today().astimezone(tz)
    to_market_close: timedelta
    try:
        to_market_close = (
            config.market_close - dt
            if config.market_close > dt
            else timedelta(hours=24) + (config.market_close - dt)
        )
        tlog(
            f"poylgon_producer tear-down task waiting for market close: {to_market_close}"
        )
    except Exception as e:
        tlog(
            f"poylgon_producer - exception of type {type(e).__name__} with args {e.args}"
        )
        return

    try:
        await asyncio.sleep(to_market_close.total_seconds() + 60 * 5)

        tlog("poylgon_producer teardown closing web-sockets")
        for w in ws:
            await w.close(False)

        tlog("poylgon_producer teardown closing tasks")

        for task in tasks:
            tlog(
                f"teardown_task(): requesting task {task.get_name()} to cancel"
            )
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                tlog("teardown_task(): task is cancelled now")

        # asyncio.get_running_loop().stop()

    except asyncio.CancelledError:
        tlog("teardown_task() cancelled during sleep")

    finally:
        tlog("teardown_task() done.")


"""
process main
"""


async def producer_async_main(
    queues: List[Queue],
    scanner_queue: Queue,
    num_consumer_processes: int,
):
    await create_db_connection(str(config.dsn))

    data_ws = tradeapi.StreamConn(
        base_url=config.prod_base_url,
        key_id=config.prod_api_key_id,
        secret_key=config.prod_api_secret,
        data_stream="polygon",
    )

    main_task = asyncio.create_task(
        run(
            data_ws=data_ws,
            queues=queues,
        ),
        name="main_task",
    )

    base_url = (
        config.prod_base_url if config.env == "PROD" else config.paper_base_url
    )
    api_key_id = (
        config.prod_api_key_id
        if config.env == "PROD"
        else config.paper_api_key_id
    )
    api_secret = (
        config.prod_api_secret
        if config.env == "PROD"
        else config.paper_api_secret
    )
    trade_ws = tradeapi.StreamConn(
        base_url=base_url,
        key_id=api_key_id,
        secret_key=api_secret,
    )

    trade_updates_task = asyncio.create_task(
        trade_run(ws=trade_ws, queues=queues),
        name="trade_updates_task",
    )

    scanner_input_task = asyncio.create_task(
        scanner_input(scanner_queue, data_ws, num_consumer_processes),
        name="scanner_input",
    )
    tear_down = asyncio.create_task(
        teardown_task(
            timezone("America/New_York"),
            [data_ws, trade_ws],
            [main_task, scanner_input_task],
        )
    )

    await asyncio.gather(
        main_task,
        trade_updates_task,
        scanner_input_task,
        tear_down,
        return_exceptions=True,
    )

    tlog("producer_async_main() completed")


def polygon_producer_main(
    unique_id: str,
    queues: List[Queue],
    current_symbols: List[str],
    current_queue_id_hash: Dict[str, int],
    market_close: datetime,
    conf_dict: Dict,
    scanner_queue: Queue,
    num_consumer_processes: int,
) -> None:
    tlog(f"*** polygon_producer_main() starting w pid {os.getpid()} ***")
    try:
        config.market_close = market_close
        config.batch_id = unique_id
        events = conf_dict.get("events", None)

        if not events:
            config.WS_DATA_CHANNELS = ["A", "AM", "T", "Q"]
        else:
            config.WS_DATA_CHANNELS = []

            if "second" in events:
                config.WS_DATA_CHANNELS.append("AM")
            if "minute" in events:
                config.WS_DATA_CHANNELS.append("A")
            if "trade" in events:
                config.WS_DATA_CHANNELS.append("T")
            if "quote" in events:
                config.WS_DATA_CHANNELS.append("Q")

        tlog(
            f"polygon_producer_main(): listening for events {config.WS_DATA_CHANNELS}"
        )
        global symbols
        global queue_id_hash

        symbols = current_symbols
        queue_id_hash = current_queue_id_hash
        if not asyncio.get_event_loop().is_closed():
            asyncio.get_event_loop().close()
        asyncio.run(
            producer_async_main(queues, scanner_queue, num_consumer_processes)
        )

    except KeyboardInterrupt:
        tlog("polygon_producer_main() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"polygon_producer_main() - exception of type {type(e).__name__} with args {e.args}"
        )

    tlog("*** polygon_producer_main() completed ***")



import os
import sys
from datetime import datetime

from google.cloud import logging

from liualgotrader.common import config

try:
    logger = logging.Client().logger("trader")
except Exception:
    logger = None


def tlog(msg: str) -> None:
    try:
        calling_fn = f"[{sys._getframe(1).f_code.co_name}()]"
    except Exception:
        calling_fn = ""

    if logger:
        try:
            logger.log_text(f"[{config.env}]{calling_fn}[{os.getpid()}] {msg}")
        except Exception as e:
            print(f"[Error] exception when trying to log to Stackdriver {e}")
            pass
    print(f"{calling_fn}[{os.getpid()}]{datetime.now()}:{msg}", flush=True)



import asyncio
import time

from liualgotrader.common.tlog import tlog


def timeit(func):
    async def process(func, *args, **params):
        if asyncio.iscoroutinefunction(func):
            return await func(*args, **params)
        else:
            return func(*args, **params)

    async def helper(*args, **params):
        tlog(f"{func.__name__} started")
        start = time.time()
        result = await process(func, *args, **params)
        tlog(f"{func.__name__} >>> {round(time.time() - start, 3)} seconds")
        return result

    return helper



from enum import Enum, auto
from multiprocessing import Queue
from typing import Awaitable, List


class WSConnectState(Enum):
    NOT_CONNECTED = auto()
    CONNECTED = auto()
    AUTHENTICATED = auto()


class StreamingBase:
    def __init__(self, queues: List[Queue]):
        self.queues = queues

    async def subscribe(self, symbol: str, handler: Awaitable) -> bool:
        pass

    async def unsubscribe(self, symbol: str) -> bool:
        pass

    async def close(
        self,
    ) -> None:
        pass



import math
from datetime import date, timedelta
from typing import Dict, List, Optional

import alpaca_trade_api as tradeapi
from talib import MAMA

from liualgotrader.common import config
from liualgotrader.common.decorators import timeit
from liualgotrader.common.market_data import \
    get_historical_data_from_polygon_by_range
from liualgotrader.common.tlog import tlog
from liualgotrader.miners.base import Miner
from liualgotrader.models.ticker_data import StockOhlc, TickerData


class DailyOHLC(Miner):
    def __init__(
        self,
        days: int,
        min_stock_price: Optional[float],
        max_stock_price: Optional[float],
        indicators: Optional[Dict],
        symbols: Optional[List[str]],
        debug=False,
    ):
        self._num_workers = 20
        self._days = days
        self._min_stock_price = min_stock_price
        self._max_stock_price = max_stock_price
        self._indicators = indicators
        self._symbols = symbols
        self._debug = debug
        self.data_api = tradeapi.REST(
            base_url=config.prod_base_url,
            key_id=config.prod_api_key_id,
            secret_key=config.prod_api_secret,
        )
        super().__init__(name="DailyOHLC")

    @property
    def days(self) -> int:
        return self._days

    @property
    def min_stock_price(self) -> Optional[float]:
        return self._min_stock_price

    @property
    def max_stock_price(self) -> Optional[float]:
        return self._max_stock_price

    @property
    def indicators(self) -> Optional[Dict]:
        return self._indicators

    @property
    def symbols(self) -> Optional[List[str]]:
        return self._symbols

    @symbols.setter
    def symbols(self, symbols: Optional[List[str]]) -> None:
        self._symbols = symbols

    @timeit
    async def load_symbol_data(
        self,
        symbol: str,
        days: int,
    ) -> None:
        start_date = date.today() - timedelta(days=days)

        _minute_data = get_historical_data_from_polygon_by_range(
            self.data_api, [symbol], start_date, "day"
        )

        if symbol in _minute_data:
            for index, row in _minute_data[symbol].iterrows():
                indicators: Dict = {}
                if self.indicators:
                    for indicator in self.indicators:
                        if indicator == "mama":
                            mama, fama = MAMA(
                                _minute_data[symbol]["close"][:index].dropna()
                            )
                            indicators["mama"] = (
                                mama[-1] if not math.isnan(mama[-1]) else None
                            )
                            indicators["fama"] = (
                                fama[-1] if not math.isnan(fama[-1]) else None
                            )

                daily_bar = StockOhlc(
                    symbol=symbol,
                    symbol_date=index,
                    open=row["open"],
                    high=row["high"],
                    low=row["low"],
                    close=row["close"],
                    volume=int(row["volume"]),
                    indicators=indicators,
                )
                await daily_bar.save()

            tlog(f"saved {len(_minute_data[symbol].index)} days for {symbol}")

    @timeit
    async def run(self) -> bool:

        if not self.symbols:
            self.symbols = await TickerData.load_symbols()

        if not self.symbols:
            return False

        # check last date
        for symbol in self.symbols:
            latest_date = await StockOhlc.get_latest_date(symbol)

            if self._debug:
                tlog(f"{symbol} latest date: {latest_date}")

            if not latest_date:
                if self._debug:
                    tlog(f"{symbol} loading {self.days} of OHLC data")
                await self.load_symbol_data(symbol, self.days)
            else:
                latest_date += timedelta(days=1)
                duration = min(self.days, (date.today() - latest_date).days)

                if self._debug:
                    tlog(f"{symbol} loading {duration} of OHLC data")
                await self.load_symbol_data(symbol, duration)

        return True



from abc import ABCMeta, abstractmethod
from datetime import timedelta, datetime
from typing import List, Optional

import alpaca_trade_api as tradeapi


class Scanner(metaclass=ABCMeta):
    def __init__(
        self,
        name: str,
        data_api: tradeapi,
        recurrence: Optional[timedelta],
        target_strategy_name: Optional[str],
        data_source: object = None,
    ):
        self.name = name
        self.recurrence = recurrence
        self.target_strategy_name = target_strategy_name
        self.data_api = data_api
        self.data_source = data_source

    @abstractmethod
    async def run(self, back_time: datetime = None) -> List[str]:
        return []

    @classmethod
    def get_supported_scanners(cls):
        return ["momentum"]



import math
from enum import Enum
from typing import Tuple

import numpy as np
import pandas as pd
import pytz
from scipy.stats import linregress

est = pytz.timezone("US/Eastern")


class SeriesTrendType(Enum):
    UNKNOWN = 0
    SHARP_DOWN = 1
    DOWN = 5
    FLAT = 10
    UP = 15
    SHARP_UP = 20


class VolatilityClassificationType(Enum):
    UNKNOWN = 0
    LOW = 1
    MEDIUM = 5
    HIGH = 10


def get_series_trend(series: pd.Series) -> Tuple[float, SeriesTrendType]:
    if len(series) < 4:
        return 0, SeriesTrendType.UNKNOWN

    length = min(10, len(series))
    try:
        np.seterr(all="raise")
        slope, _, _, _, _ = linregress(range(length), series[-length:])
        slope = round(slope, 3)
    except FloatingPointError:
        return math.inf, SeriesTrendType.UNKNOWN

    if -0.1 < slope < 0.1:
        t = SeriesTrendType.FLAT
    elif 0 < slope <= 1:
        t = SeriesTrendType.UP
    elif slope > 1:
        t = SeriesTrendType.SHARP_UP
    elif -1 <= slope < 0:
        t = SeriesTrendType.DOWN
    else:
        t = SeriesTrendType.SHARP_DOWN

    return slope, t
